"""Definitions of points defined relatively to previous constructions."""

from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.definition import JGEXDefinition, SketchConstruction, v
from newclid.jgex.sketch import (
    sketch_amirror,
    sketch_between_bound,
    sketch_bisect,
    sketch_circle,
    sketch_exbisect,
    sketch_excenter2,
    sketch_function_name,
    sketch_incenter2,
    sketch_line,
    sketch_midp,
    sketch_pmirror,
    sketch_reflect,
    sketch_shift,
    sketch_tline,
    sketch_trisect,
    sketch_trisegment,
)

ANGLE_BISECTOR = JGEXDefinition(
    name=JGEXConstructionName.ANGLE_BISECTOR,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("eqangle b a b x b x b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_bisect.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


ANGLE_MIRROR = JGEXDefinition(
    name=JGEXConstructionName.ANGLE_MIRROR,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("eqangle b a b c b c b x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_amirror.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


EQDISTANCE = JGEXDefinition(
    name=JGEXConstructionName.EQDISTANCE,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c")[0],
    clauses=JGEXClause.from_str("cong x a b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


FOOT = JGEXDefinition(
    name=JGEXConstructionName.FOOT,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("perp x a b c, coll x b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_line.__name__), args=(v("b"), v("c"))
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


INCENTER = JGEXDefinition(
    name=JGEXConstructionName.INCENTER,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "x : eqangle a b a x a x a c, eqangle c a c x c x c b; eqangle b c b x b x b a"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_bisect.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_bisect.__name__),
            args=(v("b"), v("c"), v("a")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


INCENTER2 = JGEXDefinition(
    name=JGEXConstructionName.INCENTER2,
    args=(v("x"), v("y"), v("z"), v("i"), v("a"), v("b"), v("c")),
    rely_on_points={
        v("i"): (v("a"), v("b"), v("c")),
        v("x"): (v("i"), v("b"), v("c")),
        v("y"): (v("i"), v("c"), v("a")),
        v("z"): (v("i"), v("a"), v("b")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "i : eqangle a b a i a i a c, eqangle c a c i c i c b; eqangle b c b i b i b a; x : coll x b c, perp i x b c; y : coll y c a, perp i y c a; z : coll z a b, perp i z a b; cong i x i y, cong i y i z"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_incenter2.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"), v("y"), v("z"), v("i")),
)


EXCENTER = JGEXDefinition(
    name=JGEXConstructionName.EXCENTER,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "x : eqangle a b a x a x a c, eqangle c a c x c x c b; eqangle b c b x b x b a"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_bisect.__name__),
            args=(v("b"), v("a"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_exbisect.__name__),
            args=(v("b"), v("c"), v("a")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


EXCENTER2 = JGEXDefinition(
    name=JGEXConstructionName.EXCENTER2,
    args=(v("x"), v("y"), v("z"), v("i"), v("a"), v("b"), v("c")),
    rely_on_points={
        v("i"): (v("a"), v("b"), v("c")),
        v("x"): (v("i"), v("b"), v("c")),
        v("y"): (v("i"), v("c"), v("a")),
        v("z"): (v("i"), v("a"), v("b")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "i : eqangle a b a i a i a c, eqangle c a c i c i c b; eqangle b c b i b i b a; x : coll x b c, perp i x b c; y : coll y c a, perp i y c a; z : coll z a b, perp i z a b; cong i x i y, cong i y i z"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_excenter2.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"), v("y"), v("z"), v("i")),
)


MIDPOINT = JGEXDefinition(
    name=JGEXConstructionName.MIDPOINT,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("midp x a b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_midp.__name__),
            args=(v("a"), v("b")),
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)


MIRROR = JGEXDefinition(
    name=JGEXConstructionName.MIRROR,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("coll x a b, cong b a b x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_pmirror.__name__),
            args=(v("a"), v("b")),
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)


ORTHOCENTER = JGEXDefinition(
    name=JGEXConstructionName.ORTHOCENTER,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("perp x a b c, perp x b c a; perp x c a b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("b"), v("c"), v("a")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


REFLECT = JGEXDefinition(
    name=JGEXConstructionName.REFLECT,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c, ncoll a b c")[0],
    clauses=JGEXClause.from_str("cong b a b x, cong c a c x; perp b c a x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_reflect.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


SHIFT = JGEXDefinition(
    name=JGEXConstructionName.SHIFT,
    args=(v("x"), v("b"), v("c"), v("d")),
    rely_on_points={v("x"): (v("b"), v("c"), v("d"))},
    requirements=JGEXClause.from_str("diff d b")[0],
    clauses=JGEXClause.from_str("cong x b c d, cong x c b d"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_shift.__name__),
            args=(v("d"), v("c"), v("b")),
        ),
    ),
    input_points=(v("b"), v("c"), v("d")),
    output_points=(v("x"),),
)


TRISECT = JGEXDefinition(
    name=JGEXConstructionName.TRISECT,
    args=(v("x"), v("y"), v("a"), v("b"), v("c")),
    rely_on_points={
        v("x"): (v("a"), v("b"), v("c"), v("y")),
        v("y"): (v("a"), v("b"), v("c"), v("x")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "x y : coll x a c, coll y a c, eqangle b a b x b x b y, eqangle b x b y b y b c"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_trisect.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"), v("y")),
)


TRISEGMENT = JGEXDefinition(
    name=JGEXConstructionName.TRISEGMENT,
    args=(v("x"), v("y"), v("a"), v("b")),
    rely_on_points={
        v("x"): (v("a"), v("b"), v("y")),
        v("y"): (v("a"), v("b"), v("x")),
    },
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str(
        "x y : coll x a b, coll y a b, cong x a x y, cong y x y b, rconst a x a b 1/3"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_trisegment.__name__),
            args=(v("a"), v("b")),
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"), v("y")),
)


BETWEEN_BOUND = JGEXDefinition(
    name=JGEXConstructionName.BETWEEN_BOUND,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("coll x a b, obtuse_angle a x b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_between_bound.__name__),
            args=(v("a"), v("b")),
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)


RELATIVE_TO_CONSTRUCTIONS = [
    ANGLE_BISECTOR,
    ANGLE_MIRROR,
    EQDISTANCE,
    FOOT,
    INCENTER,
    INCENTER2,
    EXCENTER,
    EXCENTER2,
    MIDPOINT,
    MIRROR,
    ORTHOCENTER,
    REFLECT,
    SHIFT,
    TRISECT,
    TRISEGMENT,
    BETWEEN_BOUND,
]



from difflib import Differ

import pytest
from newclid.jgex.clause import JGEXConstruction
from newclid.jgex.formulation import JGEXFormulation
from newclid.jgex.jgex_setup_data import (
    JGEXClauseInProof,
    JGEXSetupData,
    JGEXSetupId,
    PredicateConstructionInSetup,
)
from newclid.llm_input import (
    AuxTrainingDatapoint,
    TrainingDatapoint,
    new_problem_from_llm_aux_output,
    problem_to_llm_input,
    problem_to_llm_input_without_predicates,
)
from newclid.numerical.geometries import PointNum
from newclid.predicate_types import PredicateArgument
from newclid.predicates import predicate_from_construction
from newclid.problem import PredicateConstruction
from newclid.proof_data import (
    PredicateInProof,
    ProofData,
    ProofId,
    ProofStep,
)
from newclid.rule import Rule, RuleApplication
from newclid.symbols.points_registry import Point, PointsRegisty


def p(name: str) -> PredicateArgument:
    return PredicateArgument(name)


@pytest.mark.parametrize(
    "aux_output",
    [
        "f = on_line f a c, on_line f b d ([C4] coll a c f, [C5] coll b d f); g = on_line g a c, on_line g b d ([C6] coll a c g, [C7] coll b d g)",
        "!aux f = on_line f a c, on_line f b d ([C4] coll a c f, [C5] coll b d f); !aux g = on_line g a c, on_line g b d ([C6] coll a c g, [C7] coll b d g)",
        "f = on_line f a c, on_line f b d; g = on_line g a c, on_line g b d",
        "!aux f = on_line f a c, on_line f b d; !aux g = on_line g a c, on_line g b d",
    ],
    ids=[
        "without_aux_tag",
        "with_aux_tag",
        "no_predicates_without_aux_tag",
        "no_predicates_with_aux_tag",
    ],
)
def test_llm_aux_output_to_jgex_problem(aux_output: str) -> None:
    initial_problem = JGEXFormulation.from_text(
        "a b c = triangle a b c; d = on_tline d b a c, on_tline d c a b ? perp a d b c"
    )
    problem_with_aux = new_problem_from_llm_aux_output(
        initial_problem=initial_problem, aux_output=aux_output, aux_tag="!aux"
    )

    expected_problem_string = "a b c = triangle a b c; d = on_tline d b a c, on_tline d c a b | f = on_line f a c, on_line f b d; g = on_line g a c, on_line g b d ? perp a d b c"
    assert str(problem_with_aux) == expected_problem_string
    assert problem_with_aux != initial_problem, "The problem should be a new object"


def test_problem_to_llm_input():
    problem = JGEXFormulation.from_text(
        "a b c = triangle a b c; d = on_tline d b a c, on_tline d c a b | f = on_line f a c, on_line f b d ? perp a d b c"
    )
    actual = problem_to_llm_input(problem, aux_tag="!aux")
    expected = "a b c = triangle a b c (); d = on_tline d b a c, on_tline d c a b ([C0] perp a c b d, [C1] perp a b c d); f = on_line f a c, on_line f b d ([C2] coll a c f, [C3] coll b d f) ? [G0] perp a d b c"
    assert actual == expected


def test_problem_to_llm_input_no_predicates():
    problem = JGEXFormulation.from_text(
        "a b c = triangle a b c; d = on_tline d b a c, on_tline d c a b | f = on_line f a c, on_line f b d ? perp a d b c"
    )
    actual = problem_to_llm_input_without_predicates(problem)
    expected = "a b c = triangle a b c; d = on_tline d b a c, on_tline d c a b; f = on_line f a c, on_line f b d ? perp a d b c"
    assert actual == expected


class TestTrainingDatapoint:
    @pytest.mark.parametrize(
        "aux_points, expected_aux_io",
        [
            (
                ["e"],
                [
                    AuxTrainingDatapoint(
                        input="a b c = triangle a b c (); d = on_tline d b a c, on_tline d c a b ([C0] perp a b c d, [C1] perp a c b d) ? [G0] perp a d b c",
                        aux_output="!aux e = on_line e a c, on_line e b d ([C2] coll a c e, [C3] coll b d e)",
                    ),
                ],
            ),
            (
                ["e", "f"],
                [
                    # n-1 aux clauses
                    AuxTrainingDatapoint(
                        input="a b c = triangle a b c (); d = on_tline d b a c, on_tline d c a b ([C0] perp a b c d, [C1] perp a c b d); e = on_line e a c, on_line e b d ([C2] coll a c e, [C3] coll b d e) ? [G0] perp a d b c, [G1] perp a d c f",
                        aux_output="!aux f = on_line f a c, on_line f b d ([C4] coll a c f, [C5] coll b d f)",
                    ),
                    # n-2 aux clauses
                    AuxTrainingDatapoint(
                        input="a b c = triangle a b c (); d = on_tline d b a c, on_tline d c a b ([C0] perp a b c d, [C1] perp a c b d) ? [G0] perp a d b c, [G1] perp a d c f",
                        aux_output="!aux e = on_line e a c, on_line e b d ([C2] coll a c e, [C3] coll b d e); !aux f = on_line f a c, on_line f b d ([C4] coll a c f, [C5] coll b d f)",
                    ),
                ],
            ),
        ],
    )
    def test_aux_io_from_setup_data(
        self,
        aux_points: list[str],
        expected_aux_io: list[AuxTrainingDatapoint],
    ):
        if "f" not in aux_points:
            aux_clauses = [self.e_aux_clause]
            goals = [self.goals[0]]
        else:
            aux_clauses = [self.e_aux_clause, self.f_aux_clause]
            goals = self.goals

        setup_data = JGEXSetupData(
            setup_clauses=self.setup_clauses, aux_clauses=aux_clauses, goals=goals
        )
        aux_io = AuxTrainingDatapoint.from_setup_data_aux_combinations(
            setup_data, aux_tag="!aux"
        )

        for actual, expected in zip(aux_io, expected_aux_io):
            assert actual == expected, "\n".join(
                Differ().compare(
                    expected.model_dump_json(indent=2).splitlines(),
                    actual.model_dump_json(indent=2).splitlines(),
                )
            )

    def test_training_datapoint_from_proof_data(self):
        """Test that from_proof_data returns a datapoint for each auxiliary clause."""

        rule = Rule(
            id="r43",
            description="Orthocenter theorem",
            premises_txt=(),
            conclusions_txt=(),
        )

        # Proof: AB ⟂ CD [C0], AC ⟂ BD [C1] (r43 Orthocenter theorem)=> AD ⟂ BC [G0]

        goal_0 = predicate_from_construction(self.goals[0].construction, self.points)
        assert goal_0 is not None
        goal_0_in_proof = PredicateInProof(id=ProofId("G0"), predicate=goal_0)
        goal_1 = predicate_from_construction(self.goals[1].construction, self.points)
        assert goal_1 is not None
        goal_1_in_proof = PredicateInProof(id=ProofId("G1"), predicate=goal_1)
        proof = [
            ProofStep(
                proven_predicate=goal_0_in_proof,
                justification=RuleApplication(predicate=goal_0, rule=rule, premises=()),
                applied_on_predicates=(ProofId("C0"), ProofId("C2")),
            ),
            ProofStep(
                proven_predicate=goal_1_in_proof,
                justification=RuleApplication(predicate=goal_1, rule=rule, premises=()),
                applied_on_predicates=(ProofId("C1"), ProofId("C4")),
            ),
        ]

        proof_data = ProofData(
            points=[
                Point(name=PredicateArgument(p.name), num=p.num)
                for p in self.points.name_to_point.values()
            ],
            proven_goals=[goal_0_in_proof, goal_1_in_proof],
            unproven_goals=[],
            construction_assumptions=[],
            trivial_predicates=[],
            numerical_checks=[],
            proof_steps=proof,
        )

        training_data = TrainingDatapoint.from_proof_data_aux_combinations(
            setup_data=JGEXSetupData(
                setup_clauses=self.setup_clauses,
                aux_clauses=list(self.aux_points_to_clause.values()),
                goals=self.goals,
            ),
            proof_data=proof_data,
            aux_tag="!aux",
        )

        expected_proof_output = (
            "[G0] perp a d b c <(r43)= [C0,C2], [G1] perp a d c f <(r43)= [C1,C4]"
        )
        assert training_data.proof_output == expected_proof_output

    @pytest.fixture(autouse=True)
    def setup(self):
        self.points = PointsRegisty()
        self.points.name_to_point = {
            p("a"): Point(name=p("a"), num=PointNum(x=0, y=0)),
            p("b"): Point(name=p("b"), num=PointNum(x=1, y=0)),
            p("c"): Point(name=p("c"), num=PointNum(x=2, y=0)),
            p("d"): Point(name=p("d"), num=PointNum(x=3, y=0)),
            p("e"): Point(name=p("e"), num=PointNum(x=4, y=0)),
            p("f"): Point(name=p("f"), num=PointNum(x=5, y=0)),
        }

        # Goal: perp a d b c
        self.goals = [
            PredicateConstructionInSetup(
                id=JGEXSetupId("G0"),
                construction=PredicateConstruction.from_str("perp a d b c"),
            ),
            PredicateConstructionInSetup(
                id=JGEXSetupId("G1"),
                construction=PredicateConstruction.from_str("perp a d f c"),
            ),
        ]

        # Setup: a b c = triangle a b c; d = on_tline d b a c, on_tline d c a b
        self.setup_clauses = [
            JGEXClauseInProof(
                constructions=(JGEXConstruction.from_str("triangle a b c"),),
                predicates=(),  # Points are implicitly defined
                added_points=("a", "b", "c"),
            ),
            JGEXClauseInProof(
                constructions=(
                    JGEXConstruction.from_str("on_tline d b a c"),
                    JGEXConstruction.from_str("on_tline d c a b"),
                ),
                predicates=(
                    PredicateConstructionInSetup(
                        id=JGEXSetupId("C0"),
                        construction=PredicateConstruction.from_str("perp a b c d"),
                    ),
                    PredicateConstructionInSetup(
                        id=JGEXSetupId("C1"),
                        construction=PredicateConstruction.from_str("perp a c b d"),
                    ),
                ),
                added_points=("d",),
            ),
        ]

        # Aux: e = on_line e a c, on_line e b d; f = on_line f a c, on_line f b d
        self.e_aux_clause = JGEXClauseInProof(
            constructions=(
                JGEXConstruction.from_str("on_line e a c"),
                JGEXConstruction.from_str("on_line e b d"),
            ),
            predicates=(
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C2"),
                    construction=PredicateConstruction.from_str("coll a c e"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C3"),
                    construction=PredicateConstruction.from_str("coll b d e"),
                ),
            ),
            added_points=("e",),
        )

        self.f_aux_clause = JGEXClauseInProof(
            constructions=(
                JGEXConstruction.from_str("on_line f a c"),
                JGEXConstruction.from_str("on_line f b d"),
            ),
            predicates=(
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C4"),
                    construction=PredicateConstruction.from_str("coll a c f"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C5"),
                    construction=PredicateConstruction.from_str("coll b d f"),
                ),
            ),
            added_points=("f",),
        )
        self.aux_points_to_clause = {
            "e": self.e_aux_clause,
            "f": self.f_aux_clause,
        }



from __future__ import annotations

from dataclasses import dataclass
from typing import cast

from matplotlib import patches
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.collections import PathCollection
from matplotlib.lines import AxLine
from matplotlib.text import Annotation
from matplotlib.typing import ColorType

from newclid.draw.figure import draw_point
from newclid.draw.theme import DrawTheme
from newclid.predicate_types import PredicateArgument
from newclid.symbols.symbols_registry import SymbolsRegistry


@dataclass
class PointArtists:
    name: str
    path: ArtistInAnimation
    annotation: ArtistInAnimation

    def highlight(self, color: ColorType | None = None) -> set[Artist]:
        updated_artists: set[Artist] = set()
        if self.path.highlight(color=color):
            updated_artists.add(self.path.artist)
        if self.annotation.highlight(color=color):
            updated_artists.add(self.annotation.artist)
        return updated_artists

    def gray_out(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        if self.path.gray_out():
            updated_artists.add(self.path.artist)
        if self.annotation.gray_out():
            updated_artists.add(self.annotation.artist)
        return updated_artists

    def hide(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        if self.path.hide():
            updated_artists.add(self.path.artist)
        if self.annotation.hide():
            updated_artists.add(self.annotation.artist)
        return updated_artists


@dataclass
class ArtistInAnimation:
    artist: Artist
    original_color: ColorType | None = None
    original_zorder: float | None = None
    original_linewidth: float | None = None
    original_arrowstyle: patches.ArrowStyle | None = None

    def __post_init__(self) -> None:
        if isinstance(self.artist, AxLine):
            self.original_color = self.artist.get_color()
            self.original_linewidth = self.artist.get_linewidth()
        elif isinstance(self.artist, patches.FancyArrowPatch):
            self.original_arrowstyle = self.artist.get_arrowstyle()
        elif isinstance(self.artist, Annotation):
            self.original_color = self.artist.get_color()
        elif isinstance(self.artist, patches.Patch):
            self.original_color = self.artist.get_facecolor()
            self.original_linewidth = self.artist.get_linewidth()
        elif isinstance(self.artist, PathCollection):
            self.original_color = cast(ColorType, self.artist.get_facecolor()[0])
            self.original_linewidth = cast(float, self.artist.get_linewidth())
        if isinstance(self.artist, Annotation):
            self.original_size = self.artist.get_fontsize()

        self.original_zorder = self.artist.get_zorder()

    def highlight(self, color: ColorType | None = None) -> bool:
        was_not_visible = self.show()

        current_alpha = self.artist.get_alpha()
        is_angle = isinstance(self.artist, patches.Wedge)
        will_change_alpha = current_alpha != 1 and not is_angle
        if not is_angle:
            self.artist.set_alpha(1)
            self.artist.set_zorder(max(self.original_zorder or 10.0, 10.0))

        updated_kwargs: dict[str, str | float | ColorType] = {}

        if color is not None:
            updated_kwargs["color"] = color
        elif self.original_color is not None:
            updated_kwargs["color"] = self.original_color

        if self.original_linewidth is not None:
            updated_kwargs["linewidth"] = self.original_linewidth
        if self.original_arrowstyle is not None:
            updated_kwargs["arrowstyle"] = "simple,head_length=10,head_width=10"

        self.artist.update(updated_kwargs)
        return was_not_visible or will_change_alpha

    def gray_out(self) -> bool:
        is_visible = self.artist.get_visible()
        was_not_grayed_out = self.artist.get_alpha() != 0.5
        self.artist.set_alpha(0.5)
        self.artist.set_zorder(1.0)

        updated_kwargs: dict[str, str | float | ColorType] = {"color": "gray"}
        if self.original_linewidth is not None:
            updated_kwargs["linewidth"] = self.original_linewidth / 2
        if self.original_arrowstyle is not None:
            updated_kwargs["arrowstyle"] = "simple,head_length=0,head_width=0"

        self.artist.update(updated_kwargs)
        return is_visible and was_not_grayed_out

    def show(self) -> bool:
        was_not_visible = not self.artist.get_visible()
        self.artist.set_visible(True)
        return was_not_visible

    def hide(self) -> bool:
        was_visible = self.artist.get_visible()
        self.artist.set_visible(False)
        return was_visible


def init_point_name_to_artist(
    ax: Axes,
    color_theme: DrawTheme,
    aux_point_names: set[PredicateArgument],
    symbols: SymbolsRegistry,
) -> dict[str, PointArtists]:
    point_name_to_artist: dict[str, PointArtists] = {}
    for point in symbols.points:
        point_color = (
            color_theme.aux_point_color
            if point.name in aux_point_names
            else color_theme.point_color
        )
        point_artist, annotation_artist = draw_point(
            ax, point, point_color=point_color, fontsize=18, zorder=100
        )
        point_name_to_artist[point.name] = PointArtists(
            name=point.name,
            path=ArtistInAnimation(point_artist),
            annotation=ArtistInAnimation(annotation_artist),
        )
    return point_name_to_artist



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from newclid.numerical import close_enough
from newclid.numerical.geometries import CircleNum
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.predicates.congruence import Cong
from newclid.predicates.cyclic import Cyclic
from newclid.symbols.points_registry import Point

if TYPE_CHECKING:
    from newclid.proof_state import ProofState


class Circumcenter(PredicateInterface):
    """circle O A B C -
    Represent that O is the center of the circle through A, B, and C
    (circumcenter of triangle ABC).

    Can be equivalent to cong O A O B and cong O A O C,
    and equivalent pairs of congruences.
    """

    predicate_type: Literal[PredicateType.CIRCUMCENTER] = PredicateType.CIRCUMCENTER

    center: Point
    points: tuple[Point, ...]

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        if len(set(args)) <= 3:
            return None
        return (args[0],) + tuple(sorted(set(args[1:])))

    def check_numerical(self) -> bool:
        radius = self.center.num.distance(self.points[0].num)
        circle = CircleNum(center=self.center.num, radius=radius)
        return all(
            close_enough(circle.radius, circle.center.distance(p.num))
            for p in self.points[1:]
        )

    def check(self, proof_state: ProofState) -> bool | None:
        o = self.center
        p0 = self.points[0]
        for p1 in self.points[1:]:
            cong = Cong(segment1=(o, p0), segment2=(o, p1))
            if not proof_state.check(cong):
                return False
        return True

    def add(self, proof_state: ProofState) -> tuple[PredicateInterface, ...]:
        o = self.center
        p0 = self.points[0]
        consequences: list[PredicateInterface] = []
        for p1 in self.points[1:]:
            cong = Cong(segment1=(o, p0), segment2=(o, p1))
            consequences.append(cong)
        if len(self.points) > 3:
            cyclic = Cyclic(points=self.points[1:])
            consequences.append(cyclic)
        return tuple(consequences)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        str_tokens = [self.center.name] + [p.name for p in self.points]
        return tuple(PredicateArgument(t) for t in str_tokens)

    def __str__(self) -> str:
        points_str = "".join(str(p) for p in self.points)
        return f"{self.center} is the circumcenter of the circle ●{points_str}"



"""External API for Newclid.

This module contains the main entry point for building and using a geometric solver.
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.random import Generator as RngGenerator
from typing_extensions import Self

from newclid.agent.agents_interface import DeductiveAgent
from newclid.agent.ddarn import DDARN
from newclid.all_rules import DEFAULT_RULES
from newclid.animation import ProofAnimation
from newclid.api_defaults import APIDefault
from newclid.deductors.deductor_interface import Deductor
from newclid.deductors.sympy_ar.algebraic_manipulator import SympyARDeductor
from newclid.draw.figure import draw_figure
from newclid.draw.theme import DrawTheme
from newclid.jgex.formulation import JGEXFormulation
from newclid.predicates import Predicate
from newclid.problem import (
    PredicateConstruction,
    ProblemSetup,
    predicate_to_construction,
)
from newclid.proof_data import proof_data_from_state
from newclid.proof_state import ProofState
from newclid.proof_writing import write_proof
from newclid.rng import setup_rng
from newclid.rule import Rule, rules_from_file
from newclid.rule_matching.interface import RuleMatcher
from newclid.rule_matching.mapping_matcher import FilterMapper, MappingMatcher
from newclid.run_loop import RunInfos, run_loop
from newclid.webapp import pull_to_server

LOGGER = logging.getLogger(__name__)


class GeometricSolver:
    """External API for solving a geometric problem."""

    def __init__(
        self, proof: ProofState, rules: list[Rule], deductive_agent: DeductiveAgent
    ) -> None:
        self.proof_state = proof
        self.rules = rules
        self.rng = proof.rng
        self.deductive_agent = deductive_agent
        self.run_infos: RunInfos | None = None

    def run(self) -> bool:
        infos = run_loop(self.deductive_agent, proof=self.proof_state, rules=self.rules)
        self.run_infos = infos
        return infos.success

    def proof(
        self, goals_constructions: list[PredicateConstruction] | None = None
    ) -> str:
        if goals_constructions is None:
            goals_constructions = [
                predicate_to_construction(goal) for goal in self.proof_state.goals
            ]
        proof_data = proof_data_from_state(
            goals_constructions=goals_constructions, proof_state=self.proof_state
        )
        return write_proof(proof_data)

    def animate(
        self,
        jgex_problem: JGEXFormulation | None = None,
        theme: DrawTheme | None = None,
    ) -> FuncAnimation:
        goals_constructions = [
            predicate_to_construction(goal) for goal in self.proof_state.goals
        ]
        proof_data = proof_data_from_state(
            goals_constructions=goals_constructions, proof_state=self.proof_state
        )
        return ProofAnimation(
            proof_data=proof_data,
            symbols=self.proof_state.symbols,
            jgex_problem=jgex_problem,
            theme=theme if theme is not None else DrawTheme(),
        ).animate()

    def draw_figure(
        self,
        *,
        out_file: Optional[Path] = None,
        jgex_problem: JGEXFormulation | None,
        theme: DrawTheme | None = None,
    ) -> tuple[Figure, Axes]:
        fig, ax = plt.subplots()  # pyright: ignore
        return draw_figure(
            fig,
            ax,
            self.proof_state,
            jgex_problem=jgex_problem,
            theme=theme if theme is not None else DrawTheme(),
            save_to=out_file,
        )

    def write_all_outputs(
        self, out_folder_path: Path, jgex_problem: JGEXFormulation | None
    ):
        out_folder_path.mkdir(exist_ok=True, parents=True)
        with open(out_folder_path / "run_infos.json", "w", encoding="utf-8") as f:
            if self.run_infos is not None:
                f.write(self.run_infos.model_dump_json(indent=2))
            else:
                f.write("{}")
        with open(out_folder_path / "proof.txt", "w", encoding="utf-8") as f:
            f.write(self.proof())
        self.draw_figure(
            out_file=out_folder_path / "proof_figure.svg", jgex_problem=jgex_problem
        )
        pull_to_server(self.proof_state, server_path=out_folder_path / "html")
        LOGGER.info("Written all outputs at %s", out_folder_path)


class ProblemBuilder(ABC):
    """Interface for building a problem."""

    @abstractmethod
    def build(self) -> ProblemSetup:
        """Build a problem to be fed to the solver."""


class PythonDefault(APIDefault):
    def __init__(self, use_sympy_ar: bool = True) -> None:
        self.use_sympy_ar = use_sympy_ar

    def default_rule_matcher(self) -> RuleMatcher:
        return MappingMatcher(FilterMapper())

    def default_deductors(self) -> list[Deductor]:
        if self.use_sympy_ar:
            return [SympyARDeductor()]
        return []

    def default_deductive_agent(self) -> DeductiveAgent:
        return DDARN()

    def callback(self, proof_state: ProofState) -> None:
        pass


class GeometricSolverBuilder:
    """Main entry point for building a geometric solver.

    Adding or removing methods should be done with care as it represents the external API and should be stable.
    """

    def __init__(
        self,
        rng: RngGenerator | int | None = None,
        api_default: APIDefault | None = None,
    ) -> None:
        self.rng = setup_rng(rng)
        self.goals: list[Predicate] = []
        self.deductors: list[Deductor] | None = None
        self.deductive_agent: DeductiveAgent | None = None
        self.rule_matcher: Optional[RuleMatcher] = None
        self.rules: list[Rule] = sorted(DEFAULT_RULES, key=lambda r: r.id)
        self.api_default = (
            api_default if api_default is not None else _load_api_default()
        )

    def build(self, problem_setup: ProblemSetup) -> GeometricSolver:
        """Build a geometric solver with the current configuration on the given problem setup."""
        if (
            self.rule_matcher is None
            or self.deductors is None
            or self.deductive_agent is None
        ):
            if self.rule_matcher is None:
                self.rule_matcher = self.api_default.default_rule_matcher()
            if self.deductors is None:
                self.deductors = self.api_default.default_deductors()
            if self.deductive_agent is None:
                self.deductive_agent = self.api_default.default_deductive_agent()

        proof_state = ProofState(
            problem=problem_setup,
            rule_matcher=self.rule_matcher,
            deductors=self.deductors,
            rng=self.rng,
        )

        self.api_default.callback(proof_state)
        return GeometricSolver(proof_state, self.rules, self.deductive_agent)

    def with_deductive_agent(self, deductive_agent: DeductiveAgent) -> Self:
        self.deductive_agent = deductive_agent
        return self

    def with_rule_matcher(self, rule_matcher: RuleMatcher) -> Self:
        self.rule_matcher = rule_matcher
        return self

    def with_deductors(self, deductors: list[Deductor]) -> Self:
        self.deductors = deductors
        return self

    def with_rules(self, rules: list[Rule]) -> Self:
        self.rules = rules
        return self

    def with_rules_from_file(self, rules_path: Path) -> Self:
        self.with_rules(rules_from_file(rules_path))
        return self


def _load_api_default() -> APIDefault:
    try:
        from py_yuclid.api_default import HEDefault  # type: ignore

        return HEDefault()  # type: ignore
    except ImportError:
        return PythonDefault()  # type: ignore



from typing import TypeVar

from pydantic import BaseModel

PALETTE = [
    "#e6194b",
    "#3cb44b",
    "#ffe119",
    "#4363d8",
    "#f58231",
    "#911eb4",
    "#31a8a8",
    "#f032e6",
    "#bcf60c",
    "#fabebe",
    "#008080",
    "#e6beff",
    "#9a6324",
    "#fffac8",
    "#800000",
    "#aaffc3",
    "#808000",
    "#ffd8b1",
    "#0000cd",
    "#808080",
    "#00ff73",
    "#ffffff",
]

HARMONIC_BLUE = "#00479e"


class DrawTheme(BaseModel):
    circle_color: str = "#c63661"
    line_color: str = "#ffd1a9"
    triangle_color: str = "#B35555"
    goal_color: str = "#A38905"
    aux_point_color: str = HARMONIC_BLUE
    construction_color: str = "#487f56"
    point_color: str = "#3ee838"
    text_color: str = "#ffd1a9"
    perpendicular_color: str = "#FFE100"
    title_color: str = "#FFFFFF"
    thick_line_width: int = 3
    thin_line_width: int = 2


K1 = TypeVar("K1")
V1 = TypeVar("V1")


def fill_missing(d0: dict[K1, V1], d1: dict[K1, V1]) -> dict[K1, V1]:
    combined: dict[K1, V1] = d0.copy()  # type: ignore
    for k in d1.keys():
        if k not in d0:
            combined[k] = d1[k]
    return combined



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from newclid.justifications.justification import Justification
from newclid.rule import Rule

if TYPE_CHECKING:
    from newclid.proof_state import ProofState


class RuleMatcher(ABC):
    @abstractmethod
    def match_theorem(self, rule: Rule, proof: ProofState) -> set[Justification]:
        """Match all dependencies created by the given rule given the existing dependency graph."""



from __future__ import annotations

from typing import Iterator, Literal

from pydantic import BaseModel

from newclid.justifications._index import JustificationError, JustificationType
from newclid.numerical.geometries import CircleNum, circle_num_from_points_around
from newclid.predicates.cyclic import Cyclic
from newclid.symbols._index import SymbolType
from newclid.symbols.merging_symbols import SymbolMergeHistory, merge_symbols
from newclid.symbols.points_registry import Point


class CircleSymbol(BaseModel):
    """
    The CircleSymbol class represents a circle identitfied by a cyclic predicate in the symbolic registry. It has the following attributes:
    - points: the set of points known to lie on the circle.
    - justification: a Cyclic predicate that justifies the existence of this circle.
    - num: a CircleNum that represents the numerical parameter of the circle.
    - symbol_type: a constant indicating that this is a CircleSymbol (SymbolType.CIRCLE).
    """

    points: set[Point]
    justification: Cyclic
    num: CircleNum
    symbol_type: Literal[SymbolType.CIRCLE] = SymbolType.CIRCLE

    def __str__(self) -> str:
        sorted_points_names = sorted(p.name for p in self.points)
        return "Circle(" + "-".join(sorted_points_names) + ")"

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(str(self))


class CirclesRegistry:
    def __init__(self) -> None:
        self._circles: list[CircleSymbol] = []
        self._merge_history: dict[CircleSymbol, SymbolMergeHistory[CircleSymbol]] = {}

    def create_circle_thru_points(
        self, points: set[Point], because: Cyclic
    ) -> CircleSymbol:
        circle_num = circle_num_from_points_around([p.num for p in points])
        circle = CircleSymbol(justification=because, points=points, num=circle_num)
        self._circles.append(circle)
        self._merge_history[circle] = SymbolMergeHistory(symbol=circle)
        return circle

    def circle_containing(self, pnames: set[Point]) -> CircleSymbol | None:
        for circle in self._circles:
            if pnames <= circle.points:
                return circle
        return None

    def check_cyclic(self, points: tuple[Point, ...]) -> bool:
        s = set(points)
        for c in self._circles:
            if s <= c.points:
                return True
        return False

    def make_cyclic(self, points: tuple[Point, ...], justification: Cyclic) -> None:
        points_set = set(points)
        merged: list[CircleSymbol] = []
        for other_circle in self._circles:
            if points_set <= other_circle.points:
                # Already belong to the an existing circle
                return
            if len(points_set & other_circle.points) >= 3:
                merged.append(other_circle)
                points_set.update(other_circle.points)

        circle_to_merge = self.create_circle_thru_points(
            points_set, because=justification
        )
        merge_symbols(
            self._merge_history[circle_to_merge],
            [self._merge_history[merged_circle] for merged_circle in merged],
            self._circles,
        )

    def why_cyclic(self, cyclic: Cyclic) -> CircleMerge:
        points_set = set(cyclic.points)
        for circle in self._circles:
            if not points_set <= circle.points:
                continue
            circle_target = circle
            for _target in self._merge_history[circle].fellows:
                if points_set <= _target.points and len(_target.points) < len(
                    circle_target.points
                ):
                    circle_target = _target
            return CircleMerge(
                predicate=cyclic,
                circle=circle_target,
                direct_justification=circle_target.justification,
            )
        raise JustificationError("Could not justify why the points would be concyclic")

    def __iter__(self) -> Iterator[CircleSymbol]:
        return iter(self._circles)


class CircleMerge(BaseModel):
    predicate: Cyclic
    """The predicate that is justified by the symbols graph merge."""
    circle: CircleSymbol
    """The circle symbol that was merged."""
    direct_justification: Cyclic
    """The previous predicate that directly justifies the merge that has the predicate as a consequence."""

    dependency_type: Literal[JustificationType.CIRCLE_MERGE] = (
        JustificationType.CIRCLE_MERGE
    )

    def __hash__(self) -> int:
        return hash(self.predicate)



# mypy: ignore-errors

from __future__ import annotations

import itertools
from fractions import Fraction
from typing import Any, Literal, Optional

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Segment
from newclid.symbols.symbols_registry import SymbolsRegistry
from newclid.tools import (
    fraction_to_len,
    fraction_to_ratio,
    str_to_fraction,
)


class LengthEquation(PredicateInterface):
    """lequation a A b B c C ... k -
    Represents the equation a * A + b * B + c * C + ... = k. The coefficients (small letters) are signed rational numbers.

    The capitalized letters can represent products of lengths, k should be given as a float.

    This generates an equation that can be added to the deductor for reasoning (AR for example), or used in theorems.
    """

    predicate_type: Literal[PredicateType.LENGTH_EQUATION] = (
        PredicateType.LENGTH_EQUATION
    )
    coefficients: list[Fraction]
    lengths: list[Segment]
    constant: Fraction

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        k = args[-1]
        equation = args[:-1]

        counter = 0
        split_equation = []
        while counter < len(equation):
            term = []
            coefficient = equation[counter]
            coefficient = str_to_fraction(coefficient)

            term.append(coefficient)
            p1, p2 = sorted((equation[counter + 1], equation[counter + 2]))
            if p1 == p2:
                return None
            term.append((p1, p2))
            counter += 3
            while counter < len(equation) and equation[counter] == "*":
                p1, p2 = sorted((equation[counter + 1], equation[counter + 2]))
                if p1 == p2:
                    return None
                term.append((p1, p2))
                counter += 3
            term_to_sort = tuple(term[1:])
            sorted_term = sorted(term_to_sort)
            sorted_term_with_coeff = [term[0], sorted_term]
            split_equation.append(tuple(sorted_term_with_coeff))

        sorted_terms = sorted(
            split_equation, key=lambda term: tuple(var for var in term[1])
        )

        simplified_equation = []
        for term, group in itertools.groupby(sorted_terms, key=lambda pair: pair[1]):
            coefficient_sum = sum(pair[0] for pair in group)
            if coefficient_sum != 0:
                simplified_equation.append((coefficient_sum, term))

        if not simplified_equation:
            return None
        final_equation: list[
            Fraction | tuple[PredicateArgument, PredicateArgument] | str
        ] = []
        for coefficient, lengths in simplified_equation:
            final_equation.append(fraction_to_ratio(coefficient))
            j = 0
            for length in lengths:
                p1, p2 = length
                final_equation.append(p1)
                final_equation.append(p2)
                if j < (len(lengths) - 1):
                    final_equation.append("*")
                    j += 1

        return tuple(final_equation + [fraction_to_len(str_to_fraction(k))])

    @classmethod
    def parse(
        cls, args: tuple[PredicateArgument, ...], symbols: SymbolsRegistry
    ) -> Optional[tuple[Any, ...]]:
        preparse = cls.preparse(args)
        if not preparse:
            return None
        equation = preparse[:-1]
        parsed = []

        for i in range(0, len(equation) - 1, 3):
            if equation[i] != "*":
                coefficient = str_to_fraction(equation[i])
                points = equation[i + 1 : i + 3]
                p1, p2 = symbols.points.names2points(points)
                parsed += [coefficient, p1, p2]
            else:
                points = equation[i + 1 : i + 3]
                p1, p2 = symbols.points.names2points(points)
                parsed += ["*", p1, p2]
        k = preparse[-1]
        return tuple(parsed + [str_to_fraction(k)])

    def check_numerical(self) -> bool:
        expected_value = float(self.args[-1])
        equation = self.args[:-1]

        evaluated_expression = 0.0
        term = 0.0
        i = 0

        while i < len(equation) - 1:
            evaluated_expression += term
            term = 1.0
            coefficient = equation[i]
            coefficient = float(coefficient)
            p1 = equation[i + 1]
            p2 = equation[i + 2]
            length_value = p1.num.distance(p2.num)
            term *= coefficient * length_value
            i += 3

            while i < len(equation) - 1 and equation[i] == "*":
                p1 = equation[i + 1]
                p2 = equation[i + 2]
                length_value = p1.num.distance(p2.num)
                term *= length_value
                i += 3

        evaluated_expression += term

        return close_enough(evaluated_expression, expected_value)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        k = self.args[-1]
        equation = self.args[:-1]

        list_of_tokens: list[PredicateArgument] = []
        for i in range(0, len(equation) - 1, 3):
            if equation[i] != "*":
                coefficient = equation[i]
                coefficient = fraction_to_ratio(coefficient)
                p1 = equation[i + 1]
                p2 = equation[i + 2]
                list_of_tokens.append(coefficient)
                list_of_tokens.append(p1.name)
                list_of_tokens.append(p2.name)
            else:
                p1 = equation[i + 1]
                p2 = equation[i + 2]
                list_of_tokens.append("*")
                list_of_tokens.append(p1.name)
                list_of_tokens.append(p2.name)

        list_of_tokens.append(fraction_to_len(k))
        return tuple(list_of_tokens)

    def __str__(self) -> str:
        equation = self.args[:-1]
        k = self.args[-1]
        pretty_string = ""
        for i in range(0, len(equation) - 1, 3):
            if equation[i] != "*":
                coefficient = fraction_to_ratio(equation[i])
                p1 = equation[i + 1]
                p2 = equation[i + 2]
                pretty_string += f"{coefficient} * |{p1}{p2}| "
            else:
                p1 = equation[i + 1]
                p2 = equation[i + 2]
                pretty_string += f"* |{p1}{p2}| "
                if equation[i + 3] != "*":
                    pretty_string += " + "

        pretty_string = pretty_string[:-2]  # Remove the last ' + '

        return pretty_string + f" = {fraction_to_len(k)}"



from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.constructions.complete_figure import COMPLETE_FORM_CONSTRUCTIONS
from newclid.jgex.constructions.free import FREE_CONSTRUCTIONS
from newclid.jgex.constructions.intersections import INTERSECTION_CONSTRUCTIONS
from newclid.jgex.constructions.point_on_object import POINT_ON_OBJECT_CONSTRUCTIONS
from newclid.jgex.constructions.predicate_prescriptions import PREDICATE_PRESCRIPTIONS
from newclid.jgex.constructions.problem_specific import PROBLEM_SPECIFIC_CONSTRUCTIONS
from newclid.jgex.constructions.relative_to import RELATIVE_TO_CONSTRUCTIONS
from newclid.jgex.definition import JGEXDefinition

ALL_JGEX_CONSTRUCTIONS = (
    FREE_CONSTRUCTIONS
    + POINT_ON_OBJECT_CONSTRUCTIONS
    + INTERSECTION_CONSTRUCTIONS
    + COMPLETE_FORM_CONSTRUCTIONS
    + PREDICATE_PRESCRIPTIONS
    + PROBLEM_SPECIFIC_CONSTRUCTIONS
    + RELATIVE_TO_CONSTRUCTIONS
)

ALL_JGEX_CONSTRUCTIONS_BY_NAME: dict[JGEXConstructionName, JGEXDefinition] = {
    c.name: c for c in ALL_JGEX_CONSTRUCTIONS
}



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Annotated, Literal, cast

from pydantic import BaseModel, Field

from newclid.deductors.deductor_interface import ARDeduction
from newclid.justifications._index import JustificationType
from newclid.predicates import Predicate
from newclid.rule import RuleApplication
from newclid.symbols.circles_registry import CircleMerge
from newclid.symbols.lines_registry import LineMerge

if TYPE_CHECKING:
    from newclid.proof_state import ProofState

LOGGER = logging.getLogger(__name__)


class Assumption(BaseModel):
    predicate: Predicate
    """The predicate justified by assumption."""

    dependency_type: Literal[JustificationType.ASSUMPTION] = (
        JustificationType.ASSUMPTION
    )

    def __hash__(self) -> int:
        return hash(self.predicate)


class NumericalCheck(BaseModel):
    predicate: Predicate
    """The predicate justified by numerical check."""

    dependency_type: Literal[JustificationType.NUMERICAL_CHECK] = (
        JustificationType.NUMERICAL_CHECK
    )

    def __hash__(self) -> int:
        return hash(self.predicate)


class Reflexivity(BaseModel):
    predicate: Predicate
    """The predicate trivially true by reflexivity."""

    dependency_type: Literal[JustificationType.REFLEXIVITY] = (
        JustificationType.REFLEXIVITY
    )

    def __hash__(self) -> int:
        return hash(self.predicate)


class DirectConsequence(BaseModel):
    predicate: Predicate
    """The predicate justified by the direct consequences."""
    premises: tuple[Predicate, ...]
    """The premises of the direct consequences."""

    dependency_type: Literal[JustificationType.DIRECT_CONSEQUENCE] = (
        JustificationType.DIRECT_CONSEQUENCE
    )

    def __hash__(self) -> int:
        return hash(self.predicate)


Justification = Annotated[
    Assumption
    | NumericalCheck
    | RuleApplication
    | ARDeduction
    | CircleMerge
    | LineMerge
    | DirectConsequence
    | Reflexivity,
    Field(discriminator="dependency_type"),
]


def justify_dependency(
    dep: Justification, proof_state: ProofState
) -> tuple[Predicate, ...]:
    match dep.dependency_type:
        case (
            JustificationType.ASSUMPTION
            | JustificationType.NUMERICAL_CHECK
            | JustificationType.REFLEXIVITY
        ):
            return ()
        case JustificationType.LINE_MERGE | JustificationType.CIRCLE_MERGE:
            return (dep.direct_justification,)
        case JustificationType.RULE_APPLICATION | JustificationType.DIRECT_CONSEQUENCE:
            return dep.premises
        case JustificationType.AR_DEDUCTION:
            if dep.ar_premises is not None:
                return tuple(premise.predicate for premise in dep.ar_premises)

            ar_justification = cast(ARDeduction, proof_state.justify(dep.predicate))
            if ar_justification.ar_premises is None:
                raise ValueError(
                    f"Could not find premises for AR deduction: {dep.predicate}"
                )
            dep.ar_premises = tuple(
                sorted(set(ar_justification.ar_premises), key=lambda x: repr(x))
            )
            return tuple(premise.predicate for premise in dep.ar_premises)
    raise ValueError(f"Unknown dependency type: {dep.dependency_type}")



from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest
from newclid.agent._index import AgentName
from newclid.api import ProblemBuilder
from newclid.cli import LogLevel, NewclidOptions, parse_cli_args
from newclid.ggb.problem_builder import GeogebraProblemBuilder
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.problem import PredicateConstruction
from newclid.tools import pretty_basemodel_diff


class TestCli:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.fixture = CliFixture()

    def test_cli_parsing_jgex_problem_from_txt(self):
        self.fixture.given_input_arguments(
            "--seed",
            "123",
            "--log-level",
            "DEBUG",
            "--agent",
            "ddarn",
            "--output-dir",
            "mydir/output",
            "--saturate",
            "jgex",
            "--problem",
            "a b c = triangle a b c; h = on_tline h b a c, on_tline h c a b ? perp a h b c",
        )
        self.fixture.when_parsing_cli_args()
        self.fixture.then_options_should_be(
            seed=123,
            log_level=LogLevel.DEBUG,
            agent=AgentName.DDARN,
            output_dir=Path("mydir/output"),
            saturate=True,
        )
        self.fixture.then_should_have_jgex_problem_builder_with_problem(
            "a b c = triangle a b c; h = on_tline h b a c, on_tline h c a b ? perp a h b c"
        )

    def test_cli_parsing_jgex_problem_from_file(self):
        self.fixture.given_input_arguments(
            "--seed",
            "123",
            "--log-level",
            "DEBUG",
            "--agent",
            "ddarn",
            "--output-dir",
            "mydir/output",
            "jgex",
            "--problem-id",
            "orthocenter",
            "--file",
            "./newclid/problems_datasets/examples.txt",
        )
        self.fixture.when_parsing_cli_args()
        self.fixture.then_should_have_jgex_problem_builder_with_problem(
            "a b c = triangle a b c; h = on_tline h b a c, on_tline h c a b ? perp a h b c"
        )

    def test_cli_parsing_ggb_problem(self):
        self.fixture.given_input_arguments(
            "ggb",
            "--file",
            "mydir/output/problem.ggb",
            "--goals",
            "coll a b c",
            "para a b c d",
        )
        self.fixture.when_parsing_cli_args()
        self.fixture.then_should_have_ggb_problem(
            file="mydir/output/problem.ggb",
            goals=["coll a b c", "para a b c d"],
        )


class CliFixture:
    def __init__(self):
        self.args: list[str] = []
        self.options: NewclidOptions = NewclidOptions(
            seed=None,
            output_dir=None,
            saturate=False,
            agent=AgentName.FOLLOW_DEDUCTIONS,
            log_level=LogLevel.INFO,
        )
        self._problem_builder: ProblemBuilder | None = None

    @property
    def problem_builder(self) -> ProblemBuilder:
        if self._problem_builder is None:
            raise ValueError("Problem builder not parsed")
        return self._problem_builder

    def given_input_arguments(self, *args: str) -> None:
        self.args = list(args)

    def when_parsing_cli_args(self) -> None:
        self.options, self._problem_builder = parse_cli_args(self.args)

    def then_options_should_be(self, **kwargs: Any) -> None:
        expected = self.options.model_copy(update=kwargs)
        assert self.options == expected, pretty_basemodel_diff(self.options, expected)

    def then_should_have_jgex_problem_builder_with_problem(self, problem: str) -> None:
        assert isinstance(self.problem_builder, JGEXProblemBuilder)
        assert str(self.problem_builder.jgex_problem) == problem

    def then_should_have_ggb_problem(self, file: str, goals: list[str]) -> None:
        assert isinstance(self.problem_builder, GeogebraProblemBuilder)
        assert self.problem_builder.ggb_file_path == Path(file)
        assert self.problem_builder.goals == [
            PredicateConstruction.from_str(g) for g in goals
        ]



"""
Rules allow the engine to create new facts given that a list of other facts is true. They correspond to mathematical theorems.

The basic language of geometrical facts are the predicates with their arguments.

Each rule is a constant instance of :ref:`newclid.rule.Rule`.

For example in:

R00_PERPENDICULARS_GIVE_PARALLEL = Rule(
    id="r00",
    description="Perpendiculars give parallel",
    premises_txt=("perp A B C D", "perp C D E F", "ncoll A B E"),
    conclusions_txt=("para A B E F",),
)

The id being between r00-r42 means it was one of the rules used in the original AlphaGeometry.

"perp", "ncoll", and "para" are predicates defining geoemtric relations between geometric objects.

A, B, C, D, E, F are points used in the rule. Because the rule doesn't have an `allow_point_repetition` argument, each point will be considered unique for the rule application.


This file contains all the rules created through the development of Newclid.

The rules currently in use are the ones in the list DEFAULT_RULES below.
"""

from __future__ import annotations

from newclid.rule import Rule

__all__ = [
    "DEFAULT_RULES",
    "ALL_RULES",
    "DEACTIVATED_RULES",
    "TRIAGE_NO_RULE_NEEDED",
    "TRIAGE_LEAVE_RULE_OUT",
    "R00_PERPENDICULARS_GIVE_PARALLEL",
    "R01_DEFINITION_OF_CYCLIC",
    "R02_PARALLEL_FROM_INCLINATION",
    "R03_ARC_DETERMINES_INTERNAL_ANGLES",
    "R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC",
    "R05_SAME_ARC_SAME_CHORD",
    "R06_BASE_OF_HALF_TRIANGLE",
    "R07_THALES_THEOREM_1",
    "R08_RIGHT_TRIANGLES_COMMON_ANGLE_1",
    "R09_SUM_OF_ANGLES_OF_A_TRIANGLE",
    "R10_RATIO_CANCELLATION",
    "R11_BISECTOR_THEOREM_1",
    "R12_BISECTOR_THEOREM_2",
    "R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES",
    "R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES",
    "R15_ARC_DETERMINES_INSCRIBED_ANGLES_TANGENT",
    "R16_SAME_ARC_GIVING_TANGENT",
    "R17_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_1",
    "R18_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_2",
    "R19_HYPOTENUSE_IS_DIAMETER",
    "R20_DIAMETER_IS_HYPOTENUSE",
    "R21_CYCLIC_TRAPEZOID",
    "R22_BISECTOR_CONSTRUCTION",
    "R23_BISECTOR_IS_PERPENDICULAR",
    "R24_CYCLIC_KITE",
    "R25_DIAGONALS_OF_PARALLELOGRAM_1",
    "R26_DIAGONALS_OF_PARALLELOGRAM_2",
    "R27_THALES_THEOREM_2",
    "R28_OVERLAPPING_PARALLELS",
    "R29_MIDPOINT_IS_AN_EQRATIO",
    "R30_RIGHT_TRIANGLES_COMMON_ANGLE_2",
    "R31_DENOMINATOR_CANCELLING",
    "R34_AA_SIMILARITY_OF_TRIANGLES_DIRECT",
    "R35_AA_SIMILARITY_OF_TRIANGLES_REVERSE",
    "R36_ASA_CONGRUENCE_OF_TRIANGLES_DIRECT",
    "R37_ASA_CONGRUENCE_OF_TRIANGLES_REVERSE",
    "R41_THALES_THEOREM_3",
    "R42_THALES_THEOREM_4",
    "R43_ORTHOCENTER_THEOREM",
    "R44_PAPPUS_THEOREM",
    "R45_SIMSONS_LINE_THEOREM",
    "R46_INCENTER_THEOREM",
    "R47_CIRCUMCENTER_THEOREM",
    "R48_CENTROID_THEOREM",
    "R49_RECOGNIZE_CENTER_OF_CIRCLE_CYCLIC",
    "R50_RECOGNIZE_CENTER_OF_CYCLIC_CONG",
    "R51_MIDPOINT_SPLITS_IN_TWO",
    "R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES",
    "R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES",
    "R54_DEFINITION_OF_MIDPOINT",
    "R55_MIDPOINT_CONG_PROPERTIES",
    "R56_MIDPOINT_COLL_PROPERTIES",
    "R57_PYTHAGORAS_THEOREM",
    "R58_SAME_CHORD_SAME_ARC_1",
    "R59_SAME_CHORD_SAME_ARC_2",
    "R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT",
    "R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE",
    "R62_SAS_SIMILARITY_OF_TRIANGLES_DIRECT",
    "R63_SAS_SIMILARITY_OF_TRIANGLES_REVERSE",
    "R64_SSS_CONGRUENCE_OF_TRIANGLES_DIRECT",
    "R65_SSS_CONGRUENCE_OF_TRIANGLES_REVERSE",
    "R66_SAS_CONGRUENCE_OF_TRIANGLES_DIRECT",
    "R67_SAS_CONGRUENCE_OF_TRIANGLES_REVERSE",
    "R68_SIMILARITY_WITHOUT_SCALING_DIRECT",
    "R69_SIMILARITY_WITHOUT_SCALING_REVERSE",
    "R70_PROJECTIVE_HARMONIC_CONJUGATE",
    "R71_RESOLUTION_OF_RATIOS",
    "R72_DISASSEMBLING_A_CIRCLE",
    "R73_DEFINITION_OF_CIRCLE",
    "R74_INTERSECTION_BISECTORS",
    "R76_CENTER_DIA",
    "R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES",
    "R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES",
    "R79_MULTIPLY_CONG_BY_ONE",
    "R80_SAME_CHORD_SAME_ARC_FOUR_POINTS_1",
    "R81_SAME_CHORD_SAME_ARC_FOUR_POINTS_2",
    "R82_PARA_OF_COLL",
    "R83_BETWEEN_CONDITION_I",
    "R84_BETWEEN_CONDITION_II",
    "R85_GENERALIZED_PYTHAGORAS_I",
    "R86_GENERALIZED_PYTHAGORAS_II",
    "R87_CIRCUMCENTER_CHARACTERIZATION_I",
    "R88_CIRCUMCENTER_CHARACTERIZATION_II",
    "R89_LENGTH_OF_MEDIAN",
    "R90_PARALLELOGRAM_LAW",
    "R91_ANGLES_OF_ISO_TRAPEZOID",
    "R92_DIAMETER_CROSSES_CENTER",
]

R00_PERPENDICULARS_GIVE_PARALLEL = Rule(
    id="r00",
    description="Perpendiculars give parallel",
    premises_txt=("perp A B C D", "perp C D E F", "ncoll A B E"),
    conclusions_txt=("para A B E F",),
)

R01_DEFINITION_OF_CYCLIC = Rule(
    id="r01",
    description="Definition of cyclic",
    premises_txt=("cong O A O B", "cong O B O C", "cong O C O D"),
    conclusions_txt=("cyclic A B C D",),
)

R02_PARALLEL_FROM_INCLINATION = Rule(
    id="r02",
    description="Parallel from inclination",
    premises_txt=("eqangle A B P Q C D P Q",),
    conclusions_txt=("para A B C D",),
)

R03_ARC_DETERMINES_INTERNAL_ANGLES = Rule(
    id="r03",
    description="Arc determines internal angles",
    premises_txt=("cyclic A B P Q",),
    conclusions_txt=(
        "eqangle P A P B Q A Q B",
        "eqangle A Q A P B Q B P",
        "eqangle A B B Q A P P Q",
        "eqangle A B A P Q B Q P",
    ),
)

R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC = Rule(
    id="r04",
    description="Congruent angles are in a cyclic",
    premises_txt=("eqangle P A P B Q A Q B", "ncoll P Q A B"),
    conclusions_txt=("cyclic A B P Q",),
)

R05_SAME_ARC_SAME_CHORD = Rule(
    id="r05",
    description="Same arc same chord",
    premises_txt=("cyclic A B C P Q R", "eqangle C A C B R P R Q"),
    conclusions_txt=("cong A B P Q",),
)

R06_BASE_OF_HALF_TRIANGLE = Rule(
    id="r06",
    description="Base of half triangle",
    premises_txt=("midp E A B", "midp F A C"),
    conclusions_txt=("para E F B C",),
)

R07_THALES_THEOREM_1 = Rule(
    id="r07",
    description="Thales Theorem I",
    premises_txt=("para A B C D", "coll O A C", "coll O B D", "ncoll O A B"),
    conclusions_txt=(
        "simtri O A B O C D",
        "eqratio O A C A O B B D",
        "eqratio O C A C O D B D",
    ),
)

R08_RIGHT_TRIANGLES_COMMON_ANGLE_1 = Rule(
    id="r08",
    description="Right triangles common angle I",
    premises_txt=("perp A B C D", "perp E F G H", "npara A B E F"),
    conclusions_txt=("eqangle A B E F C D G H",),
)

R09_SUM_OF_ANGLES_OF_A_TRIANGLE = Rule(
    id="r09",
    description="Sum of angles of a triangle",
    premises_txt=(
        "eqangle A B C D M N P Q",
        "eqangle C D E F P Q R U",
        "npara A B E F",
    ),
    conclusions_txt=("eqangle A B E F M N R U",),
)

R10_RATIO_CANCELLATION = Rule(
    id="r10",
    description="Ratio cancellation",
    premises_txt=(
        "eqratio A B C D M N P Q",
        "eqratio C D E F P Q R U",
    ),
    conclusions_txt=("eqratio A B E F M N R U",),
)

R11_BISECTOR_THEOREM_1 = Rule(
    id="r11",
    description="Bisector theorem I",
    premises_txt=(
        "eqratio D B D C A B A C",
        "coll D B C",
        "ncoll A B C",
    ),
    conclusions_txt=("eqangle A B A D A D A C",),
)

R12_BISECTOR_THEOREM_2 = Rule(
    id="r12",
    description="Bisector theorem II",
    premises_txt=("eqangle A B A D A D A C", "coll D B C", "ncoll A B C"),
    conclusions_txt=("eqratio D B D C A B A C",),
)

R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES = Rule(
    id="r13",
    description="Isosceles triangle equal angles",
    premises_txt=("cong O A O B", "ncoll O A B"),
    conclusions_txt=("eqangle O A A B A B O B",),
)

R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES = Rule(
    id="r14",
    description="Equal base angles imply isosceles",
    premises_txt=("eqangle A O A B B A B O", "ncoll O A B"),
    conclusions_txt=("cong O A O B",),
)

R15_ARC_DETERMINES_INSCRIBED_ANGLES_TANGENT = Rule(
    id="r15",
    description="Arc determines inscribed angles (tangent)",
    premises_txt=("circle O A B C", "perp O A A X"),
    conclusions_txt=("eqangle A X A B C A C B",),
)

R16_SAME_ARC_GIVING_TANGENT = Rule(
    id="r16",
    description="Same arc giving tangent",
    premises_txt=("circle O A B C", "eqangle A X A B C A C B"),
    conclusions_txt=("perp O A A X",),
)

R17_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_1 = Rule(
    id="r17",
    description="Central angle vs inscribed angle I",
    premises_txt=("circle O A B C", "midp M B C"),
    conclusions_txt=("eqangle A B A C O B O M",),
)

R18_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_2 = Rule(
    id="r18",
    description="Central angle vs inscribed angle II",
    premises_txt=("circle O A B C", "coll M B C", "eqangle A B A C O B O M"),
    conclusions_txt=("midp M B C",),
)

R19_HYPOTENUSE_IS_DIAMETER = Rule(
    id="r19",
    description="Hypotenuse is diameter",
    premises_txt=("perp A B B C", "midp M A C"),
    conclusions_txt=("cong A M B M",),
)

R20_DIAMETER_IS_HYPOTENUSE = Rule(
    id="r20",
    description="Diameter is hypotenuse",
    premises_txt=("circle O A B C", "coll O A C"),
    conclusions_txt=("perp A B B C",),
)

R21_CYCLIC_TRAPEZOID = Rule(
    id="r21",
    description="Cyclic trapezoid",
    premises_txt=("cyclic A B C D", "para A B C D"),
    conclusions_txt=("eqangle A D C D C D C B",),
)

R22_BISECTOR_CONSTRUCTION = Rule(
    id="r22",
    description="Bisector construction",
    premises_txt=("midp M A B", "perp O M A B"),
    conclusions_txt=("cong O A O B",),
)

R23_BISECTOR_IS_PERPENDICULAR = Rule(
    id="r23",
    description="Bisector is perpendicular",
    premises_txt=("cong A P B P", "cong A Q B Q"),
    conclusions_txt=("perp A B P Q",),
)

R24_CYCLIC_KITE = Rule(
    id="r24",
    description="Cyclic kite",
    premises_txt=("cong A P B P", "cong A Q B Q", "cyclic A B P Q"),
    conclusions_txt=("perp P A A Q",),
)

R25_DIAGONALS_OF_PARALLELOGRAM_1 = Rule(
    id="r25",
    description="Diagonals of parallelogram I",
    premises_txt=("midp M A B", "midp M C D"),
    conclusions_txt=("para A C B D",),
)

R26_DIAGONALS_OF_PARALLELOGRAM_2 = Rule(
    id="r26",
    description="Diagonals of parallelogram II",
    premises_txt=("midp M A B", "para A C B D", "para A D B C", "ncoll A B C"),
    conclusions_txt=("midp M C D",),
)

R27_THALES_THEOREM_2 = Rule(
    id="r27",
    description="Thales theorem II",
    premises_txt=(
        "eqratio O A A C O B B D",
        "coll O A C",
        "coll O B D",
        "ncoll A B C",
        "sameside A O C B O D",
    ),
    conclusions_txt=("para A B C D",),
)

R28_OVERLAPPING_PARALLELS = Rule(
    id="r28",
    description="Overlapping parallels",
    premises_txt=("para A B A C",),
    conclusions_txt=("coll A B C",),
)

R29_MIDPOINT_IS_AN_EQRATIO = Rule(
    id="r29",
    description="Midpoint is an eqratio",
    premises_txt=("midp M A B", "midp N C D"),
    conclusions_txt=("eqratio M A A B N C C D",),
)

R30_RIGHT_TRIANGLES_COMMON_ANGLE_2 = Rule(
    id="r30",
    description="Right triangles common angle II",
    premises_txt=("eqangle A B P Q C D U V", "perp P Q U V"),
    conclusions_txt=("perp A B C D",),
)

R31_DENOMINATOR_CANCELLING = Rule(
    id="r31",
    description="Denominator cancelling",
    premises_txt=("eqratio A B P Q C D U V", "cong P Q U V"),
    conclusions_txt=("cong A B C D",),
)

#     R32_OLD_SSS_TRIANGLE_CONGRUENCE = Rule(
#     id="r32",
#     description="SSS Triangle congruence",
#     premises_txt=("cong A B P Q", "cong B C Q R", "cong C A R P", "ncoll A B C"),
#     conclusions_txt=("contri* A B C P Q R",),
# )
#     R33_OLD_SAS_TRIANGLE_CONGRUENCE = Rule(
#     id="r33",
#     description="SAS Triangle congruence",
#     premises_txt=("cong A B P Q", "cong B C Q R", "eqangle B A B C Q P Q R", "ncoll A B C"),
#     conclusions_txt=("contri* A B C P Q R",),
# )

R34_AA_SIMILARITY_OF_TRIANGLES_DIRECT = Rule(
    id="r34",
    description="AA Similarity of triangles (Direct)",
    premises_txt=(
        "eqangle B A B C Q P Q R",
        "eqangle C A C B R P R Q",
        "ncoll A B C",
        "sameclock A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("simtri A B C P Q R",),
    allow_point_repetition=True,
)

R35_AA_SIMILARITY_OF_TRIANGLES_REVERSE = Rule(
    id="r35",
    description="AA Similarity of triangles (Reverse)",
    premises_txt=(
        "eqangle B A B C Q R Q P",
        "eqangle C A C B R Q R P",
        "ncoll A B C",
        "sameclock A B C P R Q",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("simtrir A B C P Q R",),
    allow_point_repetition=True,
)

R36_ASA_CONGRUENCE_OF_TRIANGLES_DIRECT = Rule(
    id="r36",
    description="ASA Congruence of triangles (Direct)",
    premises_txt=(
        "eqangle B A B C Q P Q R",
        "eqangle C A C B R P R Q",
        "ncoll A B C",
        "cong A B P Q",
        "sameclock A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contri A B C P Q R",),
    allow_point_repetition=True,
)

R37_ASA_CONGRUENCE_OF_TRIANGLES_REVERSE = Rule(
    id="r37",
    description="ASA Congruence of triangles (Reverse)",
    premises_txt=(
        "eqangle B A B C Q R Q P",
        "eqangle C A C B R Q R P",
        "ncoll A B C",
        "cong A B P Q",
        "sameclock A B C P R Q",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contrir A B C P Q R",),
    allow_point_repetition=True,
)

#     R38_OLD_SSS_TRIANGLE_SIMILARITY = Rule(
#     id="r38",
#     description="SSS Triangle similarity",
#     premises_txt=("eqratio B A B C Q P Q R", "eqratio C A C B R P R Q", "ncoll A B C"),
#     conclusions_txt=("simtri* A B C P Q R",),
# )
#     R39_OLD_SAS_TRIANGLE_SIMILARITY = Rule(
#     id="r39",
#     description="SAS Triangle similarity",
#     premises_txt=("eqratio B A B C Q P Q R", "eqangle B A B C Q P Q R", "ncoll A B C"),
#     conclusions_txt=("simtri* A B C P Q R",),
# )
#     R40_OLD_SIMILARITY_NO_SCALING = Rule(
#     id="r40",
#     description="Similarity without scaling"
#     premises_txt=("eqratio B A B C Q P Q R", "eqratio C A C B R P R Q", "ncoll A B C", "cong A B P Q"),
#     conclusions_txt=("contri* A B C P Q R",),
# )

R41_THALES_THEOREM_3 = Rule(
    id="r41",
    description="Thales theorem III",
    premises_txt=(
        "para A B C D",
        "coll M A D",
        "coll N B C",
        "eqratio M A M D N B N C",
        "sameside M A D N B C",
    ),
    conclusions_txt=("para M N A B",),
)

R42_THALES_THEOREM_4 = Rule(
    id="r42",
    description="Thales theorem IV",
    premises_txt=(
        "para A B C D",
        "coll M A D",
        "coll N B C",
        "para M N A B",
        "ncoll A B C",
    ),
    conclusions_txt=("eqratio M A M D N B N C",),
)

R43_ORTHOCENTER_THEOREM = Rule(
    id="r43",
    description="Orthocenter theorem",
    premises_txt=("perp A B C D", "perp A C B D", "ncoll A B C"),
    conclusions_txt=("perp A D B C",),
)

R44_PAPPUS_THEOREM = Rule(
    id="r44",
    description="Pappus's theorem",
    premises_txt=(
        "coll A B C",
        "coll P Q R",
        "coll X A Q",
        "coll X P B",
        "coll Y A R",
        "coll Y P C",
        "coll Z B R",
        "coll Z C Q",
    ),
    conclusions_txt=("coll X Y Z",),
)

R45_SIMSONS_LINE_THEOREM = Rule(
    id="r45",
    description="Simson's line theorem",
    premises_txt=(
        "cyclic A B C P",
        "coll A L C",
        "perp P L A C",
        "coll M B C",
        "perp P M B C",
        "coll N A B",
        "perp P N A B",
    ),
    conclusions_txt=("coll L M N",),
)

R46_INCENTER_THEOREM = Rule(
    id="r46",
    description="Incenter theorem",
    premises_txt=("eqangle A B A X A X A C", "eqangle B A B X B X B C", "ncoll A B C"),
    conclusions_txt=("eqangle C B C X C X C A",),
)

R47_CIRCUMCENTER_THEOREM = Rule(
    id="r47",
    description="Circumcenter theorem",
    premises_txt=(
        "midp M A B",
        "perp X M A B",
        "midp N B C",
        "perp X N B C",
        "midp P C A",
    ),
    conclusions_txt=("perp X P C A",),
)

R48_CENTROID_THEOREM = Rule(
    id="r48",
    description="Centroid theorem",
    premises_txt=("midp M A B", "coll M X C", "midp N B C", "coll N X A", "midp P C A"),
    conclusions_txt=("coll X P B",),
)

R49_RECOGNIZE_CENTER_OF_CIRCLE_CYCLIC = Rule(
    id="r49",
    description="Recognize center of circle (cyclic)",
    premises_txt=("circle O A B C", "cyclic A B C D"),
    conclusions_txt=("cong O A O D",),
)

R50_RECOGNIZE_CENTER_OF_CYCLIC_CONG = Rule(
    id="r50",
    description="Recognize center of cyclic (cong)",
    premises_txt=("cyclic A B C D", "cong O A O B", "cong O C O D", "npara A B C D"),
    conclusions_txt=("cong O A O C",),
)

R51_MIDPOINT_SPLITS_IN_TWO = Rule(
    id="r51",
    description="Midpoint splits in two",
    premises_txt=("midp M A B",),
    conclusions_txt=("rconst M A A B 1/2",),
)

R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES = Rule(
    id="r52",
    description="Properties of similar triangles (Direct)",
    premises_txt=(
        "simtri A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=(
        "eqangle B A B C Q P Q R",
        "eqangle A B A C P Q P R",
        "eqangle A C B C P R Q R",
        "eqratio B A B C Q P Q R",
        "eqratio B C A C Q R P R",
    ),
    allow_point_repetition=True,
)

R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES = Rule(
    id="r53",
    description="Properties of similar triangles (Reverse)",
    premises_txt=(
        "simtrir A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=(
        "eqangle B A B C Q R Q P",
        "eqangle A B A C P R P Q",
        "eqangle A C B C Q R P R",
        "eqratio B A B C Q P Q R",
        "eqratio B C A C Q R P R",
    ),
    allow_point_repetition=True,
)

R54_DEFINITION_OF_MIDPOINT = Rule(
    id="r54",
    description="Definition of midpoint",
    premises_txt=("cong M A M B", "coll M A B"),
    conclusions_txt=("midp M A B",),
)

R55_MIDPOINT_CONG_PROPERTIES = Rule(
    id="r55",
    description="Properties of midpoint (cong)",
    premises_txt=("midp M A B",),
    conclusions_txt=("cong M A M B",),
)

R56_MIDPOINT_COLL_PROPERTIES = Rule(
    id="r56",
    description="Properties of midpoint (coll)",
    premises_txt=("midp M A B",),
    conclusions_txt=("coll M A B",),
)

R57_PYTHAGORAS_THEOREM = Rule(
    id="r57",
    description="Pythagoras theorem",
    premises_txt=("PythagoreanPremises A B C",),
    conclusions_txt=("PythagoreanConclusions A B C",),
)

R58_SAME_CHORD_SAME_ARC_1 = Rule(
    id="r58",
    description="Same chord same arc I",
    premises_txt=(
        "cyclic A B C P Q R",
        "cong A B P Q",
        "nperp A C B C",
        "sameclock C A B R P Q",
        "sameside C A B R P Q",
    ),
    conclusions_txt=("eqangle C A C B R P R Q",),
)

R59_SAME_CHORD_SAME_ARC_2 = Rule(
    id="r59",
    description="Same chord same arc II",
    premises_txt=(
        "cyclic A B C P Q R",
        "cong A B P Q",
        "nperp A C B C",
        "sameclock C B A R P Q",
        "nsameside C B A R P Q",
    ),
    conclusions_txt=("eqangle C A C B R P R Q",),
)

R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT = Rule(
    id="r60",
    description="SSS Similarity of triangles (Direct)",
    premises_txt=(
        "eqratio B A B C Q P Q R",
        "eqratio C A C B R P R Q",
        "ncoll A B C",
        "sameclock A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("simtri A B C P Q R",),
    allow_point_repetition=True,
)

R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE = Rule(
    id="r61",
    description="SSS Similarity of triangles (Reverse)",
    premises_txt=(
        "eqratio B A B C Q P Q R",
        "eqratio C A C B R P R Q",
        "ncoll A B C",
        "sameclock A B C P R Q",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("simtrir A B C P Q R",),
    allow_point_repetition=True,
)

R62_SAS_SIMILARITY_OF_TRIANGLES_DIRECT = Rule(
    id="r62",
    description="SAS Similarity of triangles (Direct)",
    premises_txt=(
        "eqratio B A B C Q P Q R",
        "eqangle B A B C Q P Q R",
        "ncoll A B C",
        "sameclock A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("simtri A B C P Q R",),
    allow_point_repetition=True,
)

R63_SAS_SIMILARITY_OF_TRIANGLES_REVERSE = Rule(
    id="r63",
    description="SAS Similarity of triangles (Reverse)",
    premises_txt=(
        "eqratio B A B C Q P Q R",
        "eqangle B A B C Q R Q P",
        "ncoll A B C",
        "sameclock A B C P R Q",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("simtrir A B C P Q R",),
    allow_point_repetition=True,
)

R64_SSS_CONGRUENCE_OF_TRIANGLES_DIRECT = Rule(
    id="r64",
    description="SSS Congruence of triangles (Direct)",
    premises_txt=(
        "cong A B P Q",
        "cong B C Q R",
        "cong C A R P",
        "ncoll A B C",
        "sameclock A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contri A B C P Q R",),
    allow_point_repetition=True,
)

R65_SSS_CONGRUENCE_OF_TRIANGLES_REVERSE = Rule(
    id="r65",
    description="SSS Congruence of triangles (Reverse)",
    premises_txt=(
        "cong A B P Q",
        "cong B C Q R",
        "cong C A R P",
        "ncoll A B C",
        "sameclock A B C P R Q",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contrir A B C P Q R",),
    allow_point_repetition=True,
)

R66_SAS_CONGRUENCE_OF_TRIANGLES_DIRECT = Rule(
    id="r66",
    description="SAS Congruence of triangles (Direct)",
    premises_txt=(
        "cong A B P Q",
        "cong B C Q R",
        "eqangle B A B C Q P Q R",
        "ncoll A B C",
        "sameclock A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contri A B C P Q R",),
    allow_point_repetition=True,
)

R67_SAS_CONGRUENCE_OF_TRIANGLES_REVERSE = Rule(
    id="r67",
    description="SAS Congruence of triangles (Reverse)",
    premises_txt=(
        "cong A B P Q",
        "cong B C Q R",
        "eqangle B A B C Q R Q P",
        "ncoll A B C",
        "sameclock A B C P R Q",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contrir A B C P Q R",),
    allow_point_repetition=True,
)

R68_SIMILARITY_WITHOUT_SCALING_DIRECT = Rule(
    id="r68",
    description="Similarity without scaling (Direct)",
    premises_txt=(
        "eqratio B A B C Q P Q R",
        "eqratio C A C B R P R Q",
        "ncoll A B C",
        "cong A B P Q",
        "sameclock A B C P Q R",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contri A B C P Q R",),
    allow_point_repetition=True,
)

R69_SIMILARITY_WITHOUT_SCALING_REVERSE = Rule(
    id="r69",
    description="Similarity without scaling (Reverse)",
    premises_txt=(
        "eqratio B A B C Q P Q R",
        "eqratio C A C B R P R Q",
        "ncoll A B C",
        "cong A B P Q",
        "sameclock A B C P R Q",
        "diff A B C",
        "diff P Q R",
    ),
    conclusions_txt=("contrir A B C P Q R",),
    allow_point_repetition=True,
)

R70_PROJECTIVE_HARMONIC_CONJUGATE = Rule(
    id="r70",
    description="Projective harmonic conjugate",
    premises_txt=(
        "coll A L M",
        "coll A K N",
        "coll B K M",
        "coll B L N",
        "coll C M N",
        "coll D K L",
        "coll A B C D",
    ),
    conclusions_txt=("eqratio A C A D B C B D",),
)

R71_RESOLUTION_OF_RATIOS = Rule(
    id="r71",
    description="Resolution of ratios",
    premises_txt=(
        "eqratio A B A C D E D F",
        "coll A B C",
        "coll D E F",
        "sameside A B C D E F",
    ),
    conclusions_txt=("eqratio A B B C D E E F",),
)

R72_DISASSEMBLING_A_CIRCLE = Rule(
    id="r72",
    description="Disassembling a circle",
    premises_txt=("circle O A B C",),
    conclusions_txt=("cong O A O B", "cong O B O C"),
)

R73_DEFINITION_OF_CIRCLE = Rule(
    id="r73",
    description="Definition of circle",
    premises_txt=("cong O A O B", "cong O B O C"),
    conclusions_txt=("circle O A B C",),
)

R74_INTERSECTION_BISECTORS = Rule(
    id="r74",
    description="Intersection of angle bisector and perpendicular bisector",
    premises_txt=(
        "eqangle C A C D C D C B",
        "cong D A D B",
        "ncoll A B C D",
        "nperp A B C D",
    ),
    conclusions_txt=("cyclic A B C D",),
)

# R75_SEGMENT_EQUIPARTITION = Rule(
#     id="r75",
#     description="Equipartition of segments",
#     premises_txt=(
#         "eqratio A B A C D E D F",
#         "coll A B C",
#         "coll D E F",
#         "sameside A B C D E F",
#     ),
#     conclusions_txt=("eqratio A B B C D E E F",),
# )

R76_CENTER_DIA = Rule(
    id="r76",
    description="Locate midpoint of hypotenuse",
    premises_txt=("cong A M B M", "coll B M C", "perp A B A C"),
    conclusions_txt=("midp M B C",),
)

R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES = Rule(
    id="r77",
    description="Properties of congruent triangles (Direct)",
    premises_txt=("contri A B C P Q R",),
    conclusions_txt=("simtri A B C P Q R", "cong A B P Q"),
    allow_point_repetition=True,
)

R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES = Rule(
    id="r78",
    description="Properties of congruent triangles (Reverse)",
    premises_txt=("contrir A B C P Q R",),
    conclusions_txt=("simtrir A B C P Q R", "cong A B P Q"),
    allow_point_repetition=True,
)

R79_MULTIPLY_CONG_BY_ONE = Rule(
    id="r79",
    description="Divide congruence equation by segment",
    premises_txt=(
        "cong A B C D",
        "diff X Y",
        "diff A B",
        "diff C D",
    ),
    conclusions_txt=("eqratio A B X Y C D X Y",),
    allow_point_repetition=True,
)

R80_SAME_CHORD_SAME_ARC_FOUR_POINTS_1 = Rule(
    id="r80",
    description="Same chord same arc III",
    premises_txt=(
        "cyclic A B P Q",
        "cong A B P Q",
        "npara A P B Q",
    ),
    conclusions_txt=("eqangle P A P B P B Q B",),
)

R81_SAME_CHORD_SAME_ARC_FOUR_POINTS_2 = Rule(
    id="r81",
    description="Same chord same arc IV",
    premises_txt=(
        "cyclic A B P Q",
        "cong A B P Q",
        "sameclock A P B A P Q",
    ),
    conclusions_txt=("eqangle P A P B Q B P B",),
)

# The python engine deduces this without a rule,
# so it isn't included in `ALL_RULES`,
# but we need it for the C++ bridge.
R82_PARA_OF_COLL = Rule(
    id="r82",
    description="Parallel from collinear",
    premises_txt=("coll A B C",),
    conclusions_txt=(
        "para A B B C",
        "para A B A C",
    ),
)

R83_BETWEEN_CONDITION_I = Rule(
    id="r83",
    description="Betweeness condition I",
    premises_txt=("coll A B C", "obtuse_angle A B C"),
    conclusions_txt=("lequation 1/1 A B 1/1 B C -1/1 A C 0",),
)

R84_BETWEEN_CONDITION_II = Rule(
    id="r84",
    description="Betweeness condition II",
    premises_txt=("lequation 1/1 A B 1/1 B C -1/1 A C 0",),
    conclusions_txt=("coll A B C",),
)

R85_GENERALIZED_PYTHAGORAS_I = Rule(
    id="r85",
    description="Generalized Pythagorean theorem I",
    premises_txt=("perp A B C D",),
    conclusions_txt=(
        "lequation 1/1 A C * A C 1/1 B D * B D -1/1 A D * A D -1/1 B C * B C 0",
    ),
)

R86_GENERALIZED_PYTHAGORAS_II = Rule(
    id="r86",
    description="Generalized Pythagorean theorem II",
    premises_txt=(
        "lequation 1/1 A C * A C 1/1 B D * B D -1/1 A D * A D -1/1 B C * B C 0",
        "ncoll A B C",
    ),
    conclusions_txt=("perp A B C D",),
)

R87_CIRCUMCENTER_CHARACTERIZATION_I = Rule(
    id="r87",
    description="Characterization of circumcenter I",
    premises_txt=(
        "cong O A O C",
        "aequation 1/1 A B B C 1/1 C A A O 90o",
    ),
    conclusions_txt=("circle O A B C",),
)

R88_CIRCUMCENTER_CHARACTERIZATION_II = Rule(
    id="r88",
    description="Characterization of circumcenter II",
    premises_txt=("circle O A B C",),
    conclusions_txt=("aequation 1/1 A B B C 1/1 C A A O 90o",),
)

R89_LENGTH_OF_MEDIAN = Rule(
    id="r89",
    description="Length of a median of a triangle",
    premises_txt=(
        "midp M A B",
        "ncoll A B C",
    ),
    conclusions_txt=(
        "lequation 2/1 A M * A M -2/1 B C * B C -2/1 A C * A C 1/1 A B * A B 0",
    ),
)

R90_PARALLELOGRAM_LAW = Rule(
    id="r90",
    description="Parallelogram law",
    premises_txt=(
        "para A B C D",
        "para A D B C",
        "ncoll A B C",
    ),
    conclusions_txt=(
        "lequation 2/1 A B * A B 2/1 B C * B C -1/1 A C * A C -1/1 B D * B D 0",
    ),
)

R91_ANGLES_OF_ISO_TRAPEZOID = Rule(
    id="r91",
    description="Equal angles in an isosceles trapezoid",
    premises_txt=(
        "cong A B C D",
        "para A D B C",
        "npara A B C D",
    ),
    conclusions_txt=("eqangle C A C B B C B D",),
)

R92_DIAMETER_CROSSES_CENTER = Rule(
    id="r92",
    description="Any diameter of a circle crosses the center",
    premises_txt=(
        "midp O A B",
        "circle O A B C",
        "cyclic A B C D",
        "cong A B C D",
    ),
    conclusions_txt=("coll O C D",),
)

ALL_RULES: list[Rule] = [
    R00_PERPENDICULARS_GIVE_PARALLEL,
    R01_DEFINITION_OF_CYCLIC,
    R02_PARALLEL_FROM_INCLINATION,
    R03_ARC_DETERMINES_INTERNAL_ANGLES,
    R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC,
    R05_SAME_ARC_SAME_CHORD,
    R06_BASE_OF_HALF_TRIANGLE,
    R07_THALES_THEOREM_1,
    R08_RIGHT_TRIANGLES_COMMON_ANGLE_1,
    R09_SUM_OF_ANGLES_OF_A_TRIANGLE,
    R10_RATIO_CANCELLATION,
    R11_BISECTOR_THEOREM_1,
    R12_BISECTOR_THEOREM_2,
    R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES,
    R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES,
    R15_ARC_DETERMINES_INSCRIBED_ANGLES_TANGENT,
    R16_SAME_ARC_GIVING_TANGENT,
    R17_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_1,
    R18_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_2,
    R19_HYPOTENUSE_IS_DIAMETER,
    R20_DIAMETER_IS_HYPOTENUSE,
    R21_CYCLIC_TRAPEZOID,
    R22_BISECTOR_CONSTRUCTION,
    R23_BISECTOR_IS_PERPENDICULAR,
    R24_CYCLIC_KITE,
    R25_DIAGONALS_OF_PARALLELOGRAM_1,
    R26_DIAGONALS_OF_PARALLELOGRAM_2,
    R27_THALES_THEOREM_2,
    R28_OVERLAPPING_PARALLELS,
    R29_MIDPOINT_IS_AN_EQRATIO,
    R30_RIGHT_TRIANGLES_COMMON_ANGLE_2,
    R31_DENOMINATOR_CANCELLING,
    R34_AA_SIMILARITY_OF_TRIANGLES_DIRECT,
    R35_AA_SIMILARITY_OF_TRIANGLES_REVERSE,
    R36_ASA_CONGRUENCE_OF_TRIANGLES_DIRECT,
    R37_ASA_CONGRUENCE_OF_TRIANGLES_REVERSE,
    R41_THALES_THEOREM_3,
    R42_THALES_THEOREM_4,
    R43_ORTHOCENTER_THEOREM,
    R44_PAPPUS_THEOREM,
    R45_SIMSONS_LINE_THEOREM,
    R46_INCENTER_THEOREM,
    R47_CIRCUMCENTER_THEOREM,
    R48_CENTROID_THEOREM,
    R49_RECOGNIZE_CENTER_OF_CIRCLE_CYCLIC,
    R50_RECOGNIZE_CENTER_OF_CYCLIC_CONG,
    R51_MIDPOINT_SPLITS_IN_TWO,
    R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
    R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES,
    R54_DEFINITION_OF_MIDPOINT,
    R55_MIDPOINT_CONG_PROPERTIES,
    R56_MIDPOINT_COLL_PROPERTIES,
    R57_PYTHAGORAS_THEOREM,
    R58_SAME_CHORD_SAME_ARC_1,
    R59_SAME_CHORD_SAME_ARC_2,
    R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT,
    R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE,
    R62_SAS_SIMILARITY_OF_TRIANGLES_DIRECT,
    R63_SAS_SIMILARITY_OF_TRIANGLES_REVERSE,
    R64_SSS_CONGRUENCE_OF_TRIANGLES_DIRECT,
    R65_SSS_CONGRUENCE_OF_TRIANGLES_REVERSE,
    R66_SAS_CONGRUENCE_OF_TRIANGLES_DIRECT,
    R67_SAS_CONGRUENCE_OF_TRIANGLES_REVERSE,
    R68_SIMILARITY_WITHOUT_SCALING_DIRECT,
    R69_SIMILARITY_WITHOUT_SCALING_REVERSE,
    R70_PROJECTIVE_HARMONIC_CONJUGATE,
    R71_RESOLUTION_OF_RATIOS,
    R72_DISASSEMBLING_A_CIRCLE,
    R73_DEFINITION_OF_CIRCLE,
    R74_INTERSECTION_BISECTORS,
    R76_CENTER_DIA,
    R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES,
    R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES,
    R79_MULTIPLY_CONG_BY_ONE,
    R80_SAME_CHORD_SAME_ARC_FOUR_POINTS_1,
    R81_SAME_CHORD_SAME_ARC_FOUR_POINTS_2,
    R82_PARA_OF_COLL,
    R83_BETWEEN_CONDITION_I,
    R84_BETWEEN_CONDITION_II,
    R85_GENERALIZED_PYTHAGORAS_I,
    R86_GENERALIZED_PYTHAGORAS_II,
    R87_CIRCUMCENTER_CHARACTERIZATION_I,
    R88_CIRCUMCENTER_CHARACTERIZATION_II,
    R89_LENGTH_OF_MEDIAN,
    R90_PARALLELOGRAM_LAW,
    R91_ANGLES_OF_ISO_TRAPEZOID,
    R92_DIAMETER_CROSSES_CENTER,
]
"""List of all rules that existed at some point in the history of Newclid"""

DEACTIVATED_RULES: set[Rule] = {
    R57_PYTHAGORAS_THEOREM,
}
"""Rules deactivated because they were problematic for current use cases."""

TRIAGE_NO_RULE_NEEDED: set[Rule] = {
    R00_PERPENDICULARS_GIVE_PARALLEL,
    R02_PARALLEL_FROM_INCLINATION,
    R08_RIGHT_TRIANGLES_COMMON_ANGLE_1,
    R09_SUM_OF_ANGLES_OF_A_TRIANGLE,
    R10_RATIO_CANCELLATION,
    R30_RIGHT_TRIANGLES_COMMON_ANGLE_2,
    R31_DENOMINATOR_CANCELLING,
    R72_DISASSEMBLING_A_CIRCLE,
    R73_DEFINITION_OF_CIRCLE,
    # R79_MULTIPLY_CONG_BY_ONE,
}
"""Rules that can be proven by the functioning of the engine and AR (no rule needed)."""

TRIAGE_LEAVE_RULE_OUT: set[Rule] = {
    R00_PERPENDICULARS_GIVE_PARALLEL,
    R01_DEFINITION_OF_CYCLIC,
    R02_PARALLEL_FROM_INCLINATION,
    R05_SAME_ARC_SAME_CHORD,
    R06_BASE_OF_HALF_TRIANGLE,
    R08_RIGHT_TRIANGLES_COMMON_ANGLE_1,
    R09_SUM_OF_ANGLES_OF_A_TRIANGLE,
    R10_RATIO_CANCELLATION,
    R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES,
    R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES,
    R15_ARC_DETERMINES_INSCRIBED_ANGLES_TANGENT,
    R16_SAME_ARC_GIVING_TANGENT,
    R17_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_1,
    R18_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_2,
    R21_CYCLIC_TRAPEZOID,
    R22_BISECTOR_CONSTRUCTION,
    R23_BISECTOR_IS_PERPENDICULAR,
    R24_CYCLIC_KITE,
    R29_MIDPOINT_IS_AN_EQRATIO,
    R30_RIGHT_TRIANGLES_COMMON_ANGLE_2,
    R31_DENOMINATOR_CANCELLING,
    R45_SIMSONS_LINE_THEOREM,
    R47_CIRCUMCENTER_THEOREM,
    R48_CENTROID_THEOREM,
    R55_MIDPOINT_CONG_PROPERTIES,
    R72_DISASSEMBLING_A_CIRCLE,
    R73_DEFINITION_OF_CIRCLE,
    R76_CENTER_DIA,
    # R79_MULTIPLY_CONG_BY_ONE,
}
"""Rules that can be proven by other rules, so are redundant."""

DEFAULT_RULES: set[Rule] = (
    set(ALL_RULES) - DEACTIVATED_RULES - TRIAGE_NO_RULE_NEEDED - TRIAGE_LEAVE_RULE_OUT
)
"""The rules effectively used on a normal run of the engine."""



from __future__ import annotations

from typing import Any, Iterable, Optional, Self, Union

import numpy as np
from pydantic import BaseModel, model_validator

from newclid.numerical import close_enough, nearly_zero, sign


class PointNum(BaseModel):
    """
    This class contains the geometric coordinates of a point, and the functions needed to manipulate points as 2D vectors on the plane.

    Its attributes are the x and y coordinates of the point, which are floats.

    It contains the methods:

    - angle: returns the angle of the point with respect to the x-axis.
    - close_enough: checks if two points are close enough to be considered equal.
    - distance: computes the distance between the point and another point, line, or circle.
    - distance2: computes the squared distance between the point and another point, line, or circle.
    - rot90: returns a new point rotated 90 degrees counterclockwise.
    - rotatea: rotates the point by a given angle in radians.
    - rotate: rotates the point by given sine and cosine values.
    - flip: reflects the point across the y-axis.
    - perpendicular_line: returns a line perpendicular to a given line through this point.
    - foot: computes the closest point from this point on a line or circle.
    - parallel_line: returns a line parallel to a given line through this point.
    - dot: computes the dot product with another point as a vector.
    - deduplicate: class method to remove duplicate points from an iterable of points.
    """

    x: float
    y: float

    def __add__(self, p: PointNum) -> PointNum:
        return PointNum(x=self.x + p.x, y=self.y + p.y)

    def __sub__(self, p: PointNum) -> PointNum:
        return PointNum(x=self.x - p.x, y=self.y - p.y)

    def __mul__(self, f: Any) -> PointNum:
        f = float(f)
        return PointNum(x=self.x * f, y=self.y * f)

    def __rmul__(self, f: Any) -> PointNum:
        return self * float(f)

    def __truediv__(self, f: Any) -> PointNum:
        f = float(f)
        return PointNum(x=self.x / f, y=self.y / f)

    def __str__(self) -> str:
        return "PointNum({},{})".format(self.x, self.y)

    def __abs__(self) -> str:
        return np.sqrt(self.dot(self))

    def angle(self) -> float:
        return np.arctan2(self.y, self.x)

    def close_enough(self, point: PointNum) -> bool:
        return close_enough(self.x, point.x) and close_enough(self.y, point.y)

    def distance(self, p: Union[PointNum, LineNum, CircleNum]) -> float:
        return np.sqrt(self.distance2(p))

    def distance2(self, p: Union[PointNum, LineNum, CircleNum]) -> float:
        if isinstance(p, LineNum):
            return p.distance(self)
        if isinstance(p, CircleNum):
            return (p.radius - self.distance(p.center)) ** 2
        dx = self.x - p.x
        dy = self.y - p.y
        dx2 = dx * dx
        dy2 = dy * dy
        return dx2 + dy2

    def rot90(self) -> PointNum:
        return PointNum(x=-self.y, y=self.x)

    def rotatea(self, ang: Any) -> PointNum:
        sinb, cosb = np.sin(ang), np.cos(ang)
        return self.rotate(sinb, cosb)

    def rotate(self, sinb: Any, cosb: Any) -> PointNum:
        x, y = self.x, self.y
        return PointNum(x=x * cosb - y * sinb, y=x * sinb + y * cosb)

    def flip(self) -> PointNum:
        return PointNum(x=-self.x, y=self.y)

    def perpendicular_line(self, line: LineNum) -> LineNum:
        return line.perpendicular_line(self)

    def foot(self, line: Union[LineNum, CircleNum]) -> PointNum:
        if isinstance(line, LineNum):
            perpendicular_line = line.perpendicular_line(self)
            intersect = line_line_intersection(
                perpendicular_line, line, ensure_point=True
            )
            return intersect[0]
        else:
            c, r = line.center, line.radius
            return c + (self - c) * r / self.distance(c)

    def parallel_line(self, line: LineNum) -> LineNum:
        return line.parallel_line(self)

    def dot(self, other: PointNum) -> float:
        res = self.x * other.x + self.y * other.y
        return 0 if nearly_zero(res) else res

    @classmethod
    def deduplicate(cls, points: Iterable[PointNum]) -> list[PointNum]:
        res: list[PointNum] = []
        for p in points:
            if all(not r.close_enough(p) for r in res):
                res.append(p)
        return res

    def __repr__(self) -> str:
        return f"PointNum({self.x}, {self.y})"


def line_num_from_points(p1: PointNum, p2: PointNum) -> LineNum:
    a, b, c = p1.y - p2.y, p2.x - p1.x, p1.x * p2.y - p2.x * p1.y
    return LineNum(coefficients=(a, b, c))


class LineNum(BaseModel):
    """
    This class contains the numerical parameters of a line defined by the equation ax + by + c = 0.

    Its only attribute is a tuple of three floats representing the coefficients (a, b, c) of the line equation.

    It contains the methods for geometric manipulations of the line:

    - parallel_line: returns a line parallel to this line through a given point.
    - perpendicular_line: returns a line perpendicular to this line through a given point.
    - distance: computes the distance from a point to the line.
    - is_parallel: checks if this line is parallel to another line.
    - is_perp: checks if this line is perpendicular to another line.
    - is_same: checks if this line is the same as another line.
    - point_at: infers if a point is on the line by its x and/or y coordinate(s).
    - diff_side: checks if two points are on different sides of the line.
    - same_side: checks if two points are on the same side of the line.
    - angle: computes the angle of the line with respect to the x-axis.
    - angle_to: computes the angle between this line and another line.
    """

    coefficients: tuple[float, float, float]  # pyright: ignore

    @model_validator(mode="after")
    def validate_coefficients(self) -> Self:
        a, b, c = self.coefficients
        # Make sure a is always positive
        # With a == 0, Assuming a = +epsilon > 0
        # Then b such that ax + by = 0 with y>0 should be negative.
        sa = sign(a)
        sb = sign(b)
        if sa == -1 or sa == 0 and sb == 1:
            a, b, c = -a, -b, -c
        if sa == 0:
            a = 0.0
        if sb == 0:
            b = 0.0
        if nearly_zero(c):
            c = 0.0

        d: float = np.sqrt(a**2 + b**2)
        self.coefficients: tuple[float, float, float] = a / d, b / d, c / d
        return self

    def __call__(self, p: PointNum) -> float:
        a, b, c = self.coefficients
        res = p.x * a + p.y * b + c
        return 0 if nearly_zero(res) else res

    def parallel_line(self, p: PointNum) -> LineNum:
        a, b, _ = self.coefficients
        return LineNum(coefficients=(a, b, -a * p.x - b * p.y))

    def perpendicular_line(self, p: PointNum) -> LineNum:
        a, b, _ = self.coefficients
        return line_num_from_points(p, p + PointNum(x=a, y=b))

    def distance(self, p: PointNum) -> float:
        return abs(self(p))

    def is_parallel(self, other: LineNum) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return close_enough(a * y, b * x)

    def is_perp(self, other: LineNum) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return close_enough(a * x, -b * y)

    def is_same(self, other: LineNum) -> bool:
        a, b, c = self.coefficients
        x, y, z = other.coefficients
        return close_enough(a * y, b * x) and close_enough(b * z, c * y)

    def point_at(
        self, x: Optional[float] = None, y: Optional[float] = None
    ) -> Optional[PointNum]:
        """Infer the point on the line by its x and/or y coordinate(s)"""
        a, b, c = self.coefficients
        # ax + by + c = 0
        if x is None and y is not None:
            if not close_enough(a, 0):
                return PointNum(x=(-c - b * y) / a, y=y)
            else:
                return None
        elif x is not None and y is None:
            if not close_enough(b, 0):
                return PointNum(x=x, y=(-c - a * x) / b)
            else:
                return None
        elif x is not None and y is not None:
            if close_enough(a * x + b * y, -c):
                return PointNum(x=x, y=y)
        return None

    def diff_side(self, p1: PointNum, p2: PointNum) -> bool:
        d1 = self(p1)
        d2 = self(p2)
        if nearly_zero(d1) or nearly_zero(d2):
            return False
        return d1 * d2 < 0

    def same_side(self, p1: PointNum, p2: PointNum) -> bool:
        d1 = self(p1)
        d2 = self(p2)
        if close_enough(d1, 0) or close_enough(d2, 0):
            return False
        return d1 * d2 > 0

    def angle(self) -> float:
        if nearly_zero(self.coefficients[1]):
            return np.pi / 2
        phead = self.point_at(x=1)
        ptail = self.point_at(x=0)
        if phead is None or ptail is None:
            raise TypeError("Cannot compute the angle if no head or tail.")
        angle = (phead - ptail).angle() % np.pi
        if close_enough(angle, np.pi):
            return 0.0
        return angle

    def angle_to(self, other: LineNum) -> float:
        return (self.angle() - other.angle()) % np.pi

    def __repr__(self) -> str:
        return f"LineNum({self.coefficients[0]}, {self.coefficients[1]}, {self.coefficients[2]})"


def circle_num_from_points_around(points: list[PointNum]) -> CircleNum:
    if len(points) < 3:
        raise ValueError("Circle needs at least three points")
    p1, p2, p3 = points[:3]
    l12 = _perpendicular_bisector(p1, p2)
    l23 = _perpendicular_bisector(p2, p3)
    (center,) = line_line_intersection(l12, l23)
    return circle_num_from_center_and_point(center, p1)


def circle_num_from_center_and_point(center: PointNum, point: PointNum) -> CircleNum:
    return CircleNum(center=center, radius=center.distance(point))


class CircleNum(BaseModel):
    """
    This class contains the numerical parameters of a circle.

    Its attributes are the center of the circle (a PointNum) and the radius (a float).
    """

    center: PointNum
    radius: float

    def __repr__(self) -> str:
        return f"CircleNum(c={self.center}, r={self.radius})"


def _perpendicular_bisector(p1: PointNum, p2: PointNum) -> LineNum:
    midpoint = (p1 + p2) * 0.5
    bisector_point = midpoint + PointNum(x=p2.y - p1.y, y=p1.x - p2.x)
    return line_num_from_points(midpoint, bisector_point)


def line_line_intersection(
    line_1: LineNum, line_2: LineNum, ensure_point: bool = False
) -> tuple[PointNum, ...]:
    a1, b1, c1 = line_1.coefficients
    a2, b2, c2 = line_2.coefficients
    # a1x + b1y + c1 = 0
    # a2x + b2y + c2 = 0
    d = a1 * b2 - a2 * b1
    if nearly_zero(d):
        if ensure_point:
            raise ValueError(f"Expected 1 intersection between lines, got {d}")
        return ()
    return (PointNum(x=(c2 * b1 - c1 * b2) / d, y=(c1 * a2 - c2 * a1) / d),)



import logging

import numpy as np
import pytest
from newclid.all_rules import (
    ALL_RULES,
    R00_PERPENDICULARS_GIVE_PARALLEL,
    R01_DEFINITION_OF_CYCLIC,
    R02_PARALLEL_FROM_INCLINATION,
    R03_ARC_DETERMINES_INTERNAL_ANGLES,
    R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC,
    R05_SAME_ARC_SAME_CHORD,
    R06_BASE_OF_HALF_TRIANGLE,
    R07_THALES_THEOREM_1,
    R08_RIGHT_TRIANGLES_COMMON_ANGLE_1,
    R09_SUM_OF_ANGLES_OF_A_TRIANGLE,
    R10_RATIO_CANCELLATION,
    R11_BISECTOR_THEOREM_1,
    R12_BISECTOR_THEOREM_2,
    R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES,
    R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES,
    R15_ARC_DETERMINES_INSCRIBED_ANGLES_TANGENT,
    R16_SAME_ARC_GIVING_TANGENT,
    R17_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_1,
    R18_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_2,
    R19_HYPOTENUSE_IS_DIAMETER,
    R20_DIAMETER_IS_HYPOTENUSE,
    R21_CYCLIC_TRAPEZOID,
    R22_BISECTOR_CONSTRUCTION,
    R23_BISECTOR_IS_PERPENDICULAR,
    R24_CYCLIC_KITE,
    R25_DIAGONALS_OF_PARALLELOGRAM_1,
    R26_DIAGONALS_OF_PARALLELOGRAM_2,
    R27_THALES_THEOREM_2,
    R28_OVERLAPPING_PARALLELS,
    R29_MIDPOINT_IS_AN_EQRATIO,
    R30_RIGHT_TRIANGLES_COMMON_ANGLE_2,
    R31_DENOMINATOR_CANCELLING,
    R34_AA_SIMILARITY_OF_TRIANGLES_DIRECT,
    R35_AA_SIMILARITY_OF_TRIANGLES_REVERSE,
    R36_ASA_CONGRUENCE_OF_TRIANGLES_DIRECT,
    R37_ASA_CONGRUENCE_OF_TRIANGLES_REVERSE,
    R41_THALES_THEOREM_3,
    R42_THALES_THEOREM_4,
    R43_ORTHOCENTER_THEOREM,
    R44_PAPPUS_THEOREM,
    R45_SIMSONS_LINE_THEOREM,
    R46_INCENTER_THEOREM,
    R47_CIRCUMCENTER_THEOREM,
    R48_CENTROID_THEOREM,
    R49_RECOGNIZE_CENTER_OF_CIRCLE_CYCLIC,
    R50_RECOGNIZE_CENTER_OF_CYCLIC_CONG,
    R51_MIDPOINT_SPLITS_IN_TWO,
    R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES,
    R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES,
    R54_DEFINITION_OF_MIDPOINT,
    R55_MIDPOINT_CONG_PROPERTIES,
    R56_MIDPOINT_COLL_PROPERTIES,
    R58_SAME_CHORD_SAME_ARC_1,
    R59_SAME_CHORD_SAME_ARC_2,
    R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT,
    R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE,
    R62_SAS_SIMILARITY_OF_TRIANGLES_DIRECT,
    R63_SAS_SIMILARITY_OF_TRIANGLES_REVERSE,
    R64_SSS_CONGRUENCE_OF_TRIANGLES_DIRECT,
    R65_SSS_CONGRUENCE_OF_TRIANGLES_REVERSE,
    R66_SAS_CONGRUENCE_OF_TRIANGLES_DIRECT,
    R67_SAS_CONGRUENCE_OF_TRIANGLES_REVERSE,
    R68_SIMILARITY_WITHOUT_SCALING_DIRECT,
    R69_SIMILARITY_WITHOUT_SCALING_REVERSE,
    R70_PROJECTIVE_HARMONIC_CONJUGATE,
    R71_RESOLUTION_OF_RATIOS,
    R72_DISASSEMBLING_A_CIRCLE,
    R73_DEFINITION_OF_CIRCLE,
    R74_INTERSECTION_BISECTORS,
    R76_CENTER_DIA,
    R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES,
    R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES,
    R79_MULTIPLY_CONG_BY_ONE,
    R80_SAME_CHORD_SAME_ARC_FOUR_POINTS_1,
    R81_SAME_CHORD_SAME_ARC_FOUR_POINTS_2,
    R82_PARA_OF_COLL,
    R83_BETWEEN_CONDITION_I,
    R84_BETWEEN_CONDITION_II,
    R85_GENERALIZED_PYTHAGORAS_I,
    R86_GENERALIZED_PYTHAGORAS_II,
    R87_CIRCUMCENTER_CHARACTERIZATION_I,
    R88_CIRCUMCENTER_CHARACTERIZATION_II,
    R89_LENGTH_OF_MEDIAN,
    R90_PARALLELOGRAM_LAW,
    R91_ANGLES_OF_ISO_TRAPEZOID,
)
from newclid.api import GeometricSolverBuilder
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.rule import Rule

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


PROBLEM_FOR_RULE: dict[Rule, str] = {
    R00_PERPENDICULARS_GIVE_PARALLEL: "a = free a; b = free b; c = free c; d = on_tline d c a b; e = free e; f = on_tline f e c d ? para a b e f",
    R01_DEFINITION_OF_CYCLIC: "o = free o; a = free a; b = eqdistance b o o a; c = eqdistance c o o b; d = eqdistance d o o c ? cyclic a b c d",
    R02_PARALLEL_FROM_INCLINATION: "p = free p; q = free q; a = free a; b = free b; c = free c; d = on_aline0 d p q a b p q c ? para a b c d",
    R03_ARC_DETERMINES_INTERNAL_ANGLES: "a = free a; b = free b; p = free p; q = on_circum q a b p ? eqangle p a p b q a q b; eqangle a q a p b q b p; eqangle a b b q a p p q; eqangle a b a p q b q p",
    R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC: "a = free a; b = free b; q = free q; p = eqangle3 p a b q a b ? cyclic a b p q",
    R05_SAME_ARC_SAME_CHORD: "a = free a; b = free b; c = free c; p = on_circum p a b c; r = on_circum r a b c; q = on_circum q a b c, on_aline q r p b c a ? cong a b p q",
    R06_BASE_OF_HALF_TRIANGLE: "a = free a; b = free b; c = free c; e = midpoint e a b; f = midpoint f a c ? para e f b c",
    R07_THALES_THEOREM_1: "a = free a; b = free b; c = free c; d = on_pline d c a b; o = on_line o a c, on_line o b d ? eqratio o a o c o b o d; eqratio o a c a o b b d; eqratio o c a c o d b d",
    R08_RIGHT_TRIANGLES_COMMON_ANGLE_1: "a = free a; b = free b; c = free c; d = on_tline d c a b; e = free e; f = free f; g = free g; h = on_tline h g e f ? eqangle a b e f c d g h",
    R09_SUM_OF_ANGLES_OF_A_TRIANGLE: "c = free c; d = free d; p = free p; q = free q; a = free a; b = free b; m = free m; n = on_aline0 n c d a b p q m; e = free e; f = free f; r = free r; u = on_aline0 u c d e f p q r ? eqangle a b e f m n r u",
    R10_RATIO_CANCELLATION: "a = free a; b = free b; c = free c; d = free d; m = free m; n = free n; p = free p; q = eqratio q a b c d m n p; e = free e; f = free f; r = free r; u = eqratio u c d e f p q r ? eqratio a b e f m n r u",
    R11_BISECTOR_THEOREM_1: "a = free a; b = free b; c = free c; d = eqratio6 d b c a b a c, on_line d b c ? eqangle a b a d a d a c",
    R12_BISECTOR_THEOREM_2: "a = free a; b = free b; d = free d; c = on_line c b d, on_aline c a d d a b ? eqratio d b d c a b a c",
    R13_ISOSCELES_TRIANGLE_EQUAL_ANGLES: "o = free o; a = free a; b = eqdistance b o o a ? eqangle o a a b a b o b",
    R14_EQUAL_BASE_ANGLES_IMPLY_ISOSCELES: "a = free a; b = free b; o = iso_triangle_vertex_angle o a b ? cong o a o b",
    R15_ARC_DETERMINES_INSCRIBED_ANGLES_TANGENT: "a = free a; b = free b; c = free c; o = circle o a b c; x = on_tline x a a o ? eqangle a x a b c a c b",
    R16_SAME_ARC_GIVING_TANGENT: "a = free a; b = free b; c = free c; o = circle o a b c; x = on_aline x a b a c b ? perp o a a x",
    R17_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_1: "a = free a; b = free b; c = free c; o = circle o a b c; m = midpoint m b c ? eqangle a b a c o b o m",
    R18_CENTRAL_ANGLE_VS_INSCRIBED_ANGLE_2: "a = free a; b = free b; c = free c; o = circle o a b c; m = on_line m b c, on_aline m o b c a b ? midp m b c",
    R19_HYPOTENUSE_IS_DIAMETER: "a = free a; b = free b; c = on_tline c b b a; m = midpoint m a c ? cong a m b m",
    R20_DIAMETER_IS_HYPOTENUSE: "o a b c = test_r20 o a b c ? perp a b b c",
    R21_CYCLIC_TRAPEZOID: "a = free a; b = free b; c = free c; d = on_pline d c a b, on_circum d a b c ? eqangle a d c d c d c b",
    R22_BISECTOR_CONSTRUCTION: "a = free a; b = free b; m = midpoint m a b; o = on_tline o m a b ? cong o a o b",
    R23_BISECTOR_IS_PERPENDICULAR: "a = free a; p = free p; q = free q; b = eqdistance b p a p, eqdistance b q a q ? perp a b p q",
    R24_CYCLIC_KITE: "a = free a; b = free b; p = iso_triangle_vertex p a b; q = iso_triangle_vertex q a b, on_circum q a b p ? perp p a a q",
    R25_DIAGONALS_OF_PARALLELOGRAM_1: "a b c d m = test_r25 a b c d m ? para a c b d",
    R26_DIAGONALS_OF_PARALLELOGRAM_2: "a = free a; b = free b; c = free c; d = on_pline d b a c, on_pline d a b c; m = midpoint m a b ? midp m c d",
    R27_THALES_THEOREM_2: "o = free o; a = free a; b = free b; c = on_line c a o; d = eqratio d o a a c o b b, on_line d o b ? para a b c d",
    R28_OVERLAPPING_PARALLELS: "a = free a; b = free b; c = on_pline0 c a b a ? coll a b c",
    R29_MIDPOINT_IS_AN_EQRATIO: "a = free a; b = free b; c = free c; d = free d; m = midpoint m a b; n = midpoint n c d ? eqratio m a a b n c c d",
    R30_RIGHT_TRIANGLES_COMMON_ANGLE_2: "p = free p; q = free q; u = free u; v = on_tline v u p q; a = free a; b = free b; c = free c; d = on_aline0 d p q a b u v c ? perp a b c d",
    R31_DENOMINATOR_CANCELLING: "p = free p; q = free q; u = free u; v = eqdistance v u p q; a = free a; b = free b; c = free c; d = eqratio d p q a b u v c ? cong a b c d",
    R34_AA_SIMILARITY_OF_TRIANGLES_DIRECT: "a = free a; b = free b; c = free c; q = free q; r = free r; p = on_aline p q r a b c, on_aline p r q a c b ? simtri a b c p q r",
    R35_AA_SIMILARITY_OF_TRIANGLES_REVERSE: "a b c = triangle a b c; q r = segment q r; p = on_aline p q r c b a , on_aline p r q b c a ? simtrir a b c p q r",
    R36_ASA_CONGRUENCE_OF_TRIANGLES_DIRECT: "a = free a; b = free b; p = free p; q = eqdistance q p a b; r = free r; c = on_aline c b a r q p, eqangle3 c a b r p q ? contri a b c p q r",
    R37_ASA_CONGRUENCE_OF_TRIANGLES_REVERSE: "a = free a; b = free b; p = free p; q = eqdistance q p a b; r = free r; c = on_aline c b a p q r, eqangle3 c a b r q p ? contrir a b c p q r",
    R41_THALES_THEOREM_3: "a = free a; b = free b; c = free c; d = on_pline d c a b; n = on_line n b c; m = eqratio6 m a d n b n c, on_line m a d ? para m n a b",
    R42_THALES_THEOREM_4: "a = free a; b = free b; c = free c; d = on_pline d c a b; m = on_line m a d; n = on_line n b c, on_pline n m a b ? eqratio m a m d n b n c",
    R43_ORTHOCENTER_THEOREM: "a b c = triangle a b c; d = on_tline d c a b, on_tline d b a c ? perp a d b c",
    R44_PAPPUS_THEOREM: "a b = segment a b; c = on_line c a b; p q = segment p q; r = on_line r p q; x = on_line x a q, on_line x p b; y = on_line y a r, on_line y p c; z = on_line z b r, on_line z c q ? coll x y z",
    R45_SIMSONS_LINE_THEOREM: "a b c = triangle a b c; p = on_circum p a b c; l = on_line l a c, on_tline l p a c; m = on_line m b c, on_tline m p b c; n = on_line n a b, on_tline n p a b ? coll l m n",
    R46_INCENTER_THEOREM: "a b c = triangle a b c; x = angle_bisector x b a c, angle_bisector x a b c ? eqangle c b c x c x c a",
    R47_CIRCUMCENTER_THEOREM: "a b c = triangle a b c; m = midpoint m a b; n = midpoint n b c; p = midpoint p c a; x = on_tline x m a b, on_tline x n b c ? perp x p c a",
    R48_CENTROID_THEOREM: "a b c = triangle a b c; m = midpoint m a b; n = midpoint n b c; p = midpoint p c a; x = on_line x m c, on_line x n a ? coll x p b",
    R49_RECOGNIZE_CENTER_OF_CIRCLE_CYCLIC: "a b c = triangle a b c; o = circle o a b c; d = on_circum d a b c ? cong o a o d",
    R50_RECOGNIZE_CENTER_OF_CYCLIC_CONG: "a = free a; b = free b; c = free c; d = on_circum d a b c; o = on_bline o a b, on_bline o c d ? cong o a o c",
    R51_MIDPOINT_SPLITS_IN_TWO: "a b = segment a b; m = midpoint m a b ? rconst m a a b 1/2",
    R52_SIMILAR_TRIANGLES_DIRECT_PROPERTIES: "a b c = triangle a b c; p q = segment p q; r = simtri r a b c p q ? eqangle b a b c q p q r; eqangle a b a c p q p r; eqangle a c b c p r q r; eqratio b a b c q p q r; eqratio b c a c q r p r",
    R53_SIMILAR_TRIANGLES_REVERSE_PROPERTIES: "a b c = triangle a b c; p q = segment p q; r = simtrir r a b c p q ? eqangle b a b c q r q p; eqangle a b a c p r p q; eqangle a c b c q r p r; eqratio b a b c q p q r; eqratio b c a c q r p r",
    R54_DEFINITION_OF_MIDPOINT: "a b = segment a b; m = on_line m a b, on_bline m a b ? midp m a b",
    R55_MIDPOINT_CONG_PROPERTIES: "a b = segment a b; m = midpoint m a b ? cong m a m b",
    R56_MIDPOINT_COLL_PROPERTIES: "a b = segment a b; m = midpoint m a b ? coll m a b",
    # R57_PYTHAGORAS_THEOREM: "a = free a; b = lconst b a 4; c = on_tline c b a b, lconst c b 3 ? lconst a c 5", TODO: Fix it and split it into two rules for forward and backward
    R58_SAME_CHORD_SAME_ARC_1: "a = free a; b = free b; c = free c; p = on_circum p a b c; q = on_circum q b c p, eqdistance q p a b; r = on_circum r c p q ? sameclock c a b r p q; sameside c a b r p q; eqangle c a c b r p r q",
    R59_SAME_CHORD_SAME_ARC_2: "a = free a; b = free b; c = free c; p = on_circum p a b c; q = on_circum q b c p, eqdistance q p a b; r = on_circum r c p q ? sameclock c b a r p q; nsameside c b a r p q; eqangle c a c b r p r q",
    R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT: "a b c = triangle a b c; p q = segment p q; r = eqratio r b a b c q p q, eqratio6 r p q c a c b ?  sameclock a b c p q r; simtri a b c p q r",
    R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE: "a b c = triangle a b c; p q = segment p q; r = eqratio r b a b c q p q, eqratio6 r p q c a c b ? sameclock a b c p r q; simtrir a b c p q r",
    R62_SAS_SIMILARITY_OF_TRIANGLES_DIRECT: "a b c = triangle a b c; p q = segment p q; r = eqratio r b a b c q p q, on_aline0 r b a b c q p q ? sameclock a b c p q r; simtri a b c p q r",
    R63_SAS_SIMILARITY_OF_TRIANGLES_REVERSE: "a b c = triangle a b c; p q = segment p q; r = eqratio r b a b c q p q, on_aline0 r b c b a q p q ? sameclock a b c p r q; simtrir a b c p q r",
    R64_SSS_CONGRUENCE_OF_TRIANGLES_DIRECT: "a b c = triangle a b c; p = free p; q = eqdistance q p a b; r = eqdistance r q b c, eqdistance r p a c ? sameclock a b c p q r; contri a b c p q r",
    R65_SSS_CONGRUENCE_OF_TRIANGLES_REVERSE: "a b c = triangle a b c; p = free p; q = eqdistance q p a b; r = eqdistance r q b c, eqdistance r p a c ? sameclock a b c p r q; contrir a b c p q r",
    R66_SAS_CONGRUENCE_OF_TRIANGLES_DIRECT: "a b c = triangle a b c; p = free p; q = eqdistance q p a b; r = eqdistance r q b c, on_aline0 r b a b c q p q ? sameclock a b c p q r; contri a b c p q r",
    R67_SAS_CONGRUENCE_OF_TRIANGLES_REVERSE: "a b c = triangle a b c; p = free p; q = eqdistance q p a b; r = eqdistance r q b c, on_aline0 r b c b a q p q ? sameclock a b c p r q; contrir a b c p q r",
    R68_SIMILARITY_WITHOUT_SCALING_DIRECT: "a b c = triangle a b c; p = free p; q = eqdistance q p a b; r = eqratio r b a b c q p q, eqratio6 r p q c a c b ? sameclock a b c p q r; contri a b c p q r",
    R69_SIMILARITY_WITHOUT_SCALING_REVERSE: "a b c = triangle a b c; p = free p; q = eqdistance q p a b; r = eqratio r b a b c q p q, eqratio6 r p q c a c b ? sameclock a b c p r q; contrir a b c p q r",
    R70_PROJECTIVE_HARMONIC_CONJUGATE: "a b = segment a b; c = on_line c a b; l = free l; m = on_line m a l; n = on_line n l b, on_line n m c; k = on_line k a n, on_line k b m; d = on_line d a b, on_line d k l ? eqratio a c a d b c b d",
    R71_RESOLUTION_OF_RATIOS: "a b = segment a b; c = on_line c a b; d e = segment d e; f = on_line f d e, eqratio f a b a c d e d ? eqratio a b b c d e e f",
    R72_DISASSEMBLING_A_CIRCLE: "a = free a; b = free b; c = free c; o = circle o a b c ? cong o a o b; cong o b o c",
    R73_DEFINITION_OF_CIRCLE: "o a = segment o a; b = eqdistance b o o a; c = eqdistance c o o b ? circle o a b c",
    R74_INTERSECTION_BISECTORS: "a b c = triangle a b c; d = on_bline d a b, angle_bisector d a c b ? cyclic a b c d",
    R76_CENTER_DIA: "a b = segment a b; c = on_tline c a a b; m = on_bline m a b, on_line m b c ? midp m b c",
    R77_CONGRUENT_TRIANGLES_DIRECT_PROPERTIES: "a b c = triangle a b c; p = free p; q r = contri q r a b c p ? simtri a b c p q r; cong a b p q",
    R78_CONGRUENT_TRIANGLES_REVERSE_PROPERTIES: "a b c = triangle a b c; p = free p; q r = contrir q r a b c p ? simtrir a b c p q r; cong a b p q",
    R79_MULTIPLY_CONG_BY_ONE: "a b = segment a b; c = free c; d = eqdistance d c a b; x y = segment x y ? eqratio a b x y c d x y",
    R80_SAME_CHORD_SAME_ARC_FOUR_POINTS_1: "p b = segment p b; a = free a; q = on_circum q a p b, eqdistance q p a b ? sameclock a p b a q p; eqangle p a p b p b q b",
    R81_SAME_CHORD_SAME_ARC_FOUR_POINTS_2: "p b = segment p b; a = free a; q = on_circum q a p b, eqdistance q p a b ? sameclock a p b a p q; eqangle p a p b q b p b",
    R82_PARA_OF_COLL: "a b = segment a b; c = on_line c a b ? para a b b c; para a b a c",
    R83_BETWEEN_CONDITION_I: "c a b = between c a b ? lequation 1/1 a c 1/1 c b -1/1 a b 0/1",
    # R84_BETWEEN_CONDITION_II:
    R85_GENERALIZED_PYTHAGORAS_I: "a b = segment a b; c = free c; d = on_tline d c a b ? lequation 1/1 a c * a c 1/1 b d * b d -1/1 a d * a d -1/1 b c * b c 0",
    # R86_GENERALIZED_PYTHAGORAS_II:
    # R87_CIRCUMCENTER_CHARACTERIZATION_I:
    R88_CIRCUMCENTER_CHARACTERIZATION_II: "a b c = triangle a b c; o = circle o a b c ? aequation 1/1 A B B C 1/1 C A A O 90o",
    R89_LENGTH_OF_MEDIAN: "a b c = triangle a b c; m = midpoint m a b ? lequation 2/1 A M * A M -2/1 B C * B C -2/1 A C * A C 1/1 A B * A B 0",
    R90_PARALLELOGRAM_LAW: "a b = segment a b; c = free c; d = on_pline d c a b, on_pline d a b c ? lequation 2/1 A B * A B 2/1 B C * B C -1/1 A C * A C -1/1 B D * B D 0",
    R91_ANGLES_OF_ISO_TRAPEZOID: "a b = segment a b; c = free a; d = on_pline d a b c, eqdistance d c a b ? eqangle c a c b b c b d",
}


EXTRA_RULES = {
    # C++ engine matches only some of combinatorial versions.
    # Need to get closure using `cyclic`
    # to ensure that the version in the problem is proved.
    R80_SAME_CHORD_SAME_ARC_FOUR_POINTS_1: [
        R03_ARC_DETERMINES_INTERNAL_ANGLES,
        R04_CONGRUENT_ANGLES_ARE_IN_A_CYCLIC,
    ],
    # C++ engine assume `simtri` instead of `eqratio`s
    R68_SIMILARITY_WITHOUT_SCALING_DIRECT: [R60_SSS_SIMILARITY_OF_TRIANGLES_DIRECT],
    # C++ engine assume `simtri` instead of `eqratio`s
    R69_SIMILARITY_WITHOUT_SCALING_REVERSE: [R61_SSS_SIMILARITY_OF_TRIANGLES_REVERSE],
}


@pytest.mark.parametrize("rule", ALL_RULES, ids=[r.id for r in ALL_RULES])
def test_rule(rule: Rule):
    """
    Test that all rules from `ALL_RULES` are solvable by the rules reported by `Yuclid`.
    """
    if rule not in PROBLEM_FOR_RULE:
        pytest.skip(f"No problem for rule {rule.fullname}")

    YUCLID_RULES: list[Rule] = pytest.importorskip(
        "py_yuclid.yuclid_adapter"
    ).YUCLID_RULES

    if rule in (
        R44_PAPPUS_THEOREM,
        R70_PROJECTIVE_HARMONIC_CONJUGATE,
        R83_BETWEEN_CONDITION_I,
        R84_BETWEEN_CONDITION_II,
        R85_GENERALIZED_PYTHAGORAS_I,
        R86_GENERALIZED_PYTHAGORAS_II,
        R87_CIRCUMCENTER_CHARACTERIZATION_I,
        R88_CIRCUMCENTER_CHARACTERIZATION_II,
        R89_LENGTH_OF_MEDIAN,
        R90_PARALLELOGRAM_LAW,
    ):
        pytest.skip(f"Rule {rule.fullname} not matched yet")
    rules: list[Rule] = []
    if rule in EXTRA_RULES:
        rules = EXTRA_RULES[rule] + [rule]
    elif rule in YUCLID_RULES:
        rules = [rule]
    else:
        rules = list(YUCLID_RULES)
    problem_txt = PROBLEM_FOR_RULE[rule]

    LOGGER.info(f"Testing rule {rule} with problem {problem_txt}")
    rng = np.random.default_rng(123)
    solver_builder = GeometricSolverBuilder(rng=rng).with_rules(rules)
    problem_builder = JGEXProblemBuilder(rng=rng).with_problem_from_txt(problem_txt)
    solver = solver_builder.build(problem_builder.build())
    success = solver.run()
    assert success



import logging
from typing import Literal

import numpy
from pydantic import BaseModel

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.jgex.clause import JGEXClause

LOGGER = logging.getLogger(__name__)


class CentersOfCyclicHeuristicConfig(BaseModel):
    heuristic_name: Literal[HeuristicName.CENTERS] = HeuristicName.CENTERS


class CentersOfCyclicHeuristic(Heuristic):
    """Add all centers of circles with three points."""

    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: numpy.random.Generator,
    ) -> list[JGEXClause]:
        circles = list(setup.circles)
        points = list(setup.points)
        new_clauses: list[JGEXClause] = []

        for _ in range(max_new_points):
            if not circles:
                break

            circle_index = rng.choice(len(circles))
            circle = circles.pop(circle_index)

            p1, p2, p3 = sorted(circle.points[:3])

            new_point = get_available_from_alphabet(points)
            if new_point is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping center addition."
                )
                continue
            points.append(new_point)
            new_clause_string = f"{new_point} = circle {new_point} {p1} {p2} {p3}"
            new_clause = JGEXClause.from_str(new_clause_string)[0]
            new_clauses.append(new_clause)

        return new_clauses
